import os
import unittest
from pathlib import Path
from tempfile import mkstemp
from unittest.mock import Mock, patch

from InquirerPy.base.control import Choice
from InquirerPy.separator import Separator

from huggingface_hub.commands.delete_cache import (
    _CANCEL_DELETION_STR,
    DeleteCacheCommand,
    _ask_for_confirmation_no_tui,
    _get_expectations_str,
    _get_tui_choices_from_scan,
    _manual_review_no_tui,
    _read_manual_review_tmp_file,
)
from huggingface_hub.utils import SoftTemporaryDirectory, capture_output

from .testing_utils import handle_injection


class TestDeleteCacheHelpers(unittest.TestCase):
    def test_get_tui_choices_from_scan_empty(self) -> None:
        choices = _get_tui_choices_from_scan(repos={}, preselected=[], sort_by=None)
        self.assertEqual(len(choices), 1)
        self.assertIsInstance(choices[0], Choice)
        self.assertEqual(choices[0].value, _CANCEL_DELETION_STR)
        self.assertTrue(len(choices[0].name) != 0)  # Something displayed to the user
        self.assertFalse(choices[0].enabled)

    def test_get_tui_choices_from_scan_with_preselection(self) -> None:
        choices = _get_tui_choices_from_scan(
            repos=_get_cache_mock().repos,
            preselected=[
                "dataset_revision_hash_id",  # dataset_1 is preselected
                "a_revision_id_that_does_not_exist",  # unknown but will not complain
                "older_hash_id",  # only the oldest revision from model_2
            ],
            sort_by=None,  # Don't sort to maintain original order
        )
        self.assertEqual(len(choices), 8)

        # Item to cancel everything
        self.assertIsInstance(choices[0], Choice)
        self.assertEqual(choices[0].value, _CANCEL_DELETION_STR)
        self.assertTrue(len(choices[0].name) != 0)
        self.assertFalse(choices[0].enabled)

        # Dataset repo separator
        self.assertIsInstance(choices[1], Separator)
        self.assertEqual(choices[1]._line, "\nDataset dummy_dataset (8M, used 2 weeks ago)")

        # Only revision of `dummy_dataset`
        self.assertIsInstance(choices[2], Choice)
        self.assertEqual(choices[2].value, "dataset_revision_hash_id")
        self.assertEqual(
            choices[2].name,
            # truncated hash id + detached + last modified
            "dataset_: (detached) # modified 1 day ago",
        )
        self.assertTrue(choices[2].enabled)  # preselected

        # Model `dummy_model` separator
        self.assertIsInstance(choices[3], Separator)
        self.assertEqual(choices[3]._line, "\nModel dummy_model (1.4K, used 2 years ago)")

        # Recent revision of `dummy_model` (appears first due to sorting by last_modified)
        self.assertIsInstance(choices[4], Choice)
        self.assertEqual(choices[4].value, "recent_hash_id")
        self.assertEqual(choices[4].name, "recent_h: main # modified 2 years ago")
        self.assertFalse(choices[4].enabled)

        # Oldest revision of `dummy_model`
        self.assertIsInstance(choices[5], Choice)
        self.assertEqual(choices[5].value, "older_hash_id")
        self.assertEqual(choices[5].name, "older_ha: (detached) # modified 3 years ago")
        self.assertTrue(choices[5].enabled)  # preselected

        # Model `gpt2` separator
        self.assertIsInstance(choices[6], Separator)
        self.assertEqual(choices[6]._line, "\nModel gpt2 (3.6G, used 2 hours ago)")

        # Only revision of `gpt2`
        self.assertIsInstance(choices[7], Choice)
        self.assertEqual(choices[7].value, "abcdef123456789")
        self.assertEqual(choices[7].name, "abcdef12: main, refs/pr/1 # modified 2 years ago")
        self.assertFalse(choices[7].enabled)

    def test_get_tui_choices_from_scan_with_sort_size(self) -> None:
        """Test sorting by size."""
        choices = _get_tui_choices_from_scan(repos=_get_cache_mock().repos, preselected=[], sort_by="size")

        # Verify repo order: gpt2 (3.6G) -> dummy_dataset (8M) -> dummy_model (1.4K)
        self.assertIsInstance(choices[1], Separator)
        self.assertIn("gpt2", choices[1]._line)

        self.assertIsInstance(choices[3], Separator)
        self.assertIn("dummy_dataset", choices[3]._line)

        self.assertIsInstance(choices[5], Separator)
        self.assertIn("dummy_model", choices[5]._line)

    def test_get_expectations_str_on_no_deletion_item(self) -> None:
        """Test `_get_instructions` when `_CANCEL_DELETION_STR` is passed."""
        self.assertEqual(
            _get_expectations_str(
                hf_cache_info=Mock(),
                selected_hashes=["hash_1", _CANCEL_DELETION_STR, "hash_2"],
            ),
            "Nothing will be deleted.",
        )

    def test_get_expectations_str_with_selection(self) -> None:
        """Test `_get_instructions` with 2 revisions selected."""
        strategy_mock = Mock()
        strategy_mock.expected_freed_size_str = "5.1M"

        cache_mock = Mock()
        cache_mock.delete_revisions.return_value = strategy_mock

        self.assertEqual(
            _get_expectations_str(
                hf_cache_info=cache_mock,
                selected_hashes=["hash_1", "hash_2"],
            ),
            "2 revisions selected counting for 5.1M.",
        )
        cache_mock.delete_revisions.assert_called_once_with("hash_1", "hash_2")

    def test_read_manual_review_tmp_file(self) -> None:
        """Test `_read_manual_review_tmp_file`."""

        with SoftTemporaryDirectory() as tmp_dir:
            tmp_path = Path(tmp_dir) / "file.txt"

            with tmp_path.open("w") as f:
                f.writelines(
                    [
                        "# something commented out\n",
                        "###\n",
                        "\n\n\n\n",  # some empty lines
                        "    # Something commented out after spaces\n",
                        "a_revision_hash\n",
                        "a_revision_hash_with_a_comment # 2 years ago\n",
                        "    a_revision_hash_after_spaces\n",
                        "  a_revision_hash_with_a_comment_after_spaces # 2years ago\n",
                        "    # hash_commented_out #  2 years ago\n",
                        "a_revision_hash\n",  # Duplicate
                        "",  # empty line
                    ]
                )

            # Only non-commented lines are returned
            # Order is kept and lines are not de-duplicated
            self.assertListEqual(
                _read_manual_review_tmp_file(tmp_path),
                [
                    "a_revision_hash",
                    "a_revision_hash_with_a_comment",
                    "a_revision_hash_after_spaces",
                    "a_revision_hash_with_a_comment_after_spaces",
                    "a_revision_hash",
                ],
            )

    @patch("huggingface_hub.commands.delete_cache.input")
    @patch("huggingface_hub.commands.delete_cache.mkstemp")
    def test_manual_review_no_tui(self, mock_mkstemp: Mock, mock_input: Mock) -> None:
        # Mock file creation so that we know the file location in test
        fd, tmp_path = mkstemp()
        mock_mkstemp.return_value = fd, tmp_path

        # Mock cache
        cache_mock = _get_cache_mock()

        # Mock input from user
        def _input_answers():
            self.assertTrue(os.path.isfile(tmp_path))  # not deleted yet
            with open(tmp_path) as f:
                content = f.read()
            self.assertTrue(content.startswith("# INSTRUCTIONS"))

            # older_hash_id is not commented
            self.assertIn("\n    older_hash_id # Refs: (detached)", content)
            # same for abcdef123456789
            self.assertIn("\n    abcdef123456789 # Refs: main, refs/pr/1", content)
            # dataset revision is not preselected
            self.assertIn("#    dataset_revision_hash_id", content)
            # same for recent_hash_id
            self.assertIn("#    recent_hash_id", content)

            # Select dataset revision
            content = content.replace("#    dataset_revision_hash_id", "dataset_revision_hash_id")
            # Deselect abcdef123456789
            content = content.replace("abcdef123456789", "# abcdef123456789")
            with open(tmp_path, "w") as f:
                f.write(content)

            yield "no"  # User edited the file and want to see the strategy diff
            yield "y"  # User confirms

        mock_input.side_effect = _input_answers()

        # Run manual review
        with capture_output() as output:
            selected_hashes = _manual_review_no_tui(
                hf_cache_info=cache_mock, preselected=["abcdef123456789", "older_hash_id"], sort_by=None
            )

        # Tmp file has been created but is now deleted
        mock_mkstemp.assert_called_once_with(suffix=".txt")
        self.assertFalse(os.path.isfile(tmp_path))  # now deleted

        # User changed the selection
        self.assertListEqual(selected_hashes, ["dataset_revision_hash_id", "older_hash_id"])

        # Check printed instructions
        printed = output.getvalue()
        self.assertTrue(printed.startswith("TUI is disabled. In order to"))  # ...
        self.assertIn(tmp_path, printed)

        # Check input called twice
        self.assertEqual(mock_input.call_count, 2)

    @patch("huggingface_hub.commands.delete_cache.input")
    def test_ask_for_confirmation_no_tui(self, mock_input: Mock) -> None:
        """Test `_ask_for_confirmation_no_tui`."""
        # Answer yes
        mock_input.side_effect = ("y",)
        value = _ask_for_confirmation_no_tui("custom message 1", default=True)
        mock_input.assert_called_with("custom message 1 (Y/n) ")
        self.assertTrue(value)

        # Answer no
        mock_input.side_effect = ("NO",)
        value = _ask_for_confirmation_no_tui("custom message 2", default=True)
        mock_input.assert_called_with("custom message 2 (Y/n) ")
        self.assertFalse(value)

        # Answer invalid, then default
        mock_input.side_effect = ("foo", "")
        with capture_output() as output:
            value = _ask_for_confirmation_no_tui("custom message 3", default=False)
        mock_input.assert_called_with("custom message 3 (y/N) ")
        self.assertFalse(value)
        self.assertEqual(
            output.getvalue(),
            "Invalid input. Must be one of ('y', 'yes', '1', 'n', 'no', '0', '')\n",
        )

    def test_get_tui_choices_from_scan_with_different_sorts(self) -> None:
        """Test different sorting modes."""
        cache_mock = _get_cache_mock()

        # Test size sorting (largest first) - order: gpt2 (3.6G) -> dummy_dataset (8M) -> dummy_model (1.4K)
        size_choices = _get_tui_choices_from_scan(cache_mock.repos, [], sort_by="size")
        # Separators at positions 1, 3, 5
        self.assertIsInstance(size_choices[1], Separator)
        self.assertIn("gpt2", size_choices[1]._line)
        self.assertIsInstance(size_choices[3], Separator)
        self.assertIn("dummy_dataset", size_choices[3]._line)
        self.assertIsInstance(size_choices[5], Separator)
        self.assertIn("dummy_model", size_choices[5]._line)

        # Test alphabetical sorting - order: dummy_dataset -> dummy_model -> gpt2
        alpha_choices = _get_tui_choices_from_scan(cache_mock.repos, [], sort_by="alphabetical")
        # Separators at positions 1, 3, 6 (dummy_model has 2 revisions)
        self.assertIsInstance(alpha_choices[1], Separator)
        self.assertIn("dummy_dataset", alpha_choices[1]._line)
        self.assertIsInstance(alpha_choices[3], Separator)
        self.assertIn("dummy_model", alpha_choices[3]._line)
        self.assertIsInstance(alpha_choices[6], Separator)
        self.assertIn("gpt2", alpha_choices[6]._line)

        # Test lastUpdated sorting - order: dummy_dataset (1 day) -> gpt2 (2 years) -> dummy_model (3 years)
        updated_choices = _get_tui_choices_from_scan(cache_mock.repos, [], sort_by="lastUpdated")
        # Separators at positions 1, 3, 5
        self.assertIsInstance(updated_choices[1], Separator)
        self.assertIn("dummy_dataset", updated_choices[1]._line)
        self.assertIsInstance(updated_choices[3], Separator)
        self.assertIn("gpt2", updated_choices[3]._line)
        self.assertIsInstance(updated_choices[5], Separator)
        self.assertIn("dummy_model", updated_choices[5]._line)

        # Test lastUsed sorting - order: gpt2 (2h) -> dummy_dataset (2w) -> dummy_model (2y)
        used_choices = _get_tui_choices_from_scan(cache_mock.repos, [], sort_by="lastUsed")
        # Separators at positions 1, 3, 5
        self.assertIsInstance(used_choices[1], Separator)
        self.assertIn("gpt2", used_choices[1]._line)
        self.assertIsInstance(used_choices[3], Separator)
        self.assertIn("dummy_dataset", used_choices[3]._line)
        self.assertIsInstance(used_choices[5], Separator)
        self.assertIn("dummy_model", used_choices[5]._line)


@patch("huggingface_hub.commands.delete_cache._ask_for_confirmation_no_tui")
@patch("huggingface_hub.commands.delete_cache._get_expectations_str")
@patch("huggingface_hub.commands.delete_cache.inquirer.confirm")
@patch("huggingface_hub.commands.delete_cache._manual_review_tui")
@patch("huggingface_hub.commands.delete_cache._manual_review_no_tui")
@patch("huggingface_hub.commands.delete_cache.scan_cache_dir")
@handle_injection
class TestMockedDeleteCacheCommand(unittest.TestCase):
    """Test case with a patched `DeleteCacheCommand` to test `.run()` without testing
    the manual review.
    """

    args: Mock
    command: DeleteCacheCommand

    def setUp(self) -> None:
        self.args = Mock()
        self.args.sort = None
        self.command = DeleteCacheCommand(self.args)

    def test_run_and_delete_with_tui(
        self,
        mock_scan_cache_dir: Mock,
        mock__manual_review_tui: Mock,
        mock__get_expectations_str: Mock,
        mock_confirm: Mock,
    ) -> None:
        """Test command run with a mocked manual review step."""
        # Mock return values
        mock__manual_review_tui.return_value = ["hash_1", "hash_2"]
        mock__get_expectations_str.return_value = "Will delete A and B."
        mock_confirm.return_value.execute.return_value = True
        mock_scan_cache_dir.return_value = _get_cache_mock()

        # Run
        self.command.disable_tui = False
        with capture_output() as output:
            self.command.run()

        # Step 1: scan
        mock_scan_cache_dir.assert_called_once_with(self.args.dir)
        cache_mock = mock_scan_cache_dir.return_value

        # Step 2: manual review
        mock__manual_review_tui.assert_called_once_with(cache_mock, preselected=[], sort_by=None)

        # Step 3: ask confirmation
        mock__get_expectations_str.assert_called_once_with(cache_mock, ["hash_1", "hash_2"])
        mock_confirm.assert_called_once_with("Will delete A and B. Confirm deletion ?", default=True)
        mock_confirm().execute.assert_called_once_with()

        # Step 4: delete
        cache_mock.delete_revisions.assert_called_once_with("hash_1", "hash_2")
        strategy_mock = cache_mock.delete_revisions.return_value
        strategy_mock.execute.assert_called_once_with()

        # Check output
        assert "Start deletion.\nDone. Deleted 0 repo(s) and 0 revision(s) for a total of 5.1M.\n" in output.getvalue()

    def test_run_nothing_selected_with_tui(self, mock__manual_review_tui: Mock) -> None:
        """Test command run but nothing is selected in manual review."""
        # Mock return value
        mock__manual_review_tui.return_value = []

        # Run
        self.command.disable_tui = False
        with capture_output() as output:
            self.command.run()

        # Check output
        assert "Deletion is cancelled. Do nothing.\n" in output.getvalue()

    def test_run_stuff_selected_but_cancel_item_as_well_with_tui(self, mock__manual_review_tui: Mock) -> None:
        """Test command run when some are selected but "cancel item" as well."""
        # Mock return value
        mock__manual_review_tui.return_value = [
            "hash_1",
            "hash_2",
            _CANCEL_DELETION_STR,
        ]

        # Run
        self.command.disable_tui = False
        with capture_output() as output:
            self.command.run()

        # Check output
        assert "Deletion is cancelled. Do nothing.\n" in output.getvalue()

    def test_run_and_delete_no_tui(
        self,
        mock_scan_cache_dir: Mock,
        mock__manual_review_no_tui: Mock,
        mock__get_expectations_str: Mock,
        mock__ask_for_confirmation_no_tui: Mock,
    ) -> None:
        """Test command run with a mocked manual review step."""
        # Mock return values
        mock__manual_review_no_tui.return_value = ["hash_1", "hash_2"]
        mock__get_expectations_str.return_value = "Will delete A and B."
        mock__ask_for_confirmation_no_tui.return_value.return_value = True
        mock_scan_cache_dir.return_value = _get_cache_mock()

        # Run
        self.command.disable_tui = True
        with capture_output() as output:
            self.command.run()

        # Step 1: scan
        mock_scan_cache_dir.assert_called_once_with(self.args.dir)
        cache_mock = mock_scan_cache_dir.return_value

        # Step 2: manual review
        mock__manual_review_no_tui.assert_called_once_with(cache_mock, preselected=[], sort_by=None)

        # Step 3: ask confirmation
        mock__get_expectations_str.assert_called_once_with(cache_mock, ["hash_1", "hash_2"])
        mock__ask_for_confirmation_no_tui.assert_called_once_with("Will delete A and B. Confirm deletion ?")

        # Step 4: delete
        cache_mock.delete_revisions.assert_called_once_with("hash_1", "hash_2")
        strategy_mock = cache_mock.delete_revisions.return_value
        strategy_mock.execute.assert_called_once_with()

        # Check output
        assert "Start deletion.\nDone. Deleted 0 repo(s) and 0 revision(s) for a total of 5.1M.\n" in output.getvalue()

    def test_run_with_sorting(self):
        """Test command run with sorting enabled."""
        self.args.sort = "size"
        self.command = DeleteCacheCommand(self.args)

        mock_scan_cache_dir = Mock()
        mock_scan_cache_dir.return_value = _get_cache_mock()

        with patch("huggingface_hub.commands.delete_cache.scan_cache_dir", mock_scan_cache_dir):
            with patch("huggingface_hub.commands.delete_cache._manual_review_tui") as mock_review:
                self.command.disable_tui = False
                self.command.run()

                mock_review.assert_called_once_with(mock_scan_cache_dir.return_value, preselected=[], sort_by="size")


def _get_cache_mock() -> Mock:
    # First model with 1 revision
    model_1 = Mock()
    model_1.repo_type = "model"
    model_1.repo_id = "gpt2"
    model_1.size_on_disk_str = "3.6G"
    model_1.last_accessed = 1660000000
    model_1.last_accessed_str = "2 hours ago"
    model_1.size_on_disk = 3.6 * 1024**3  # 3.6 GiB

    model_1_revision_1 = Mock()
    model_1_revision_1.commit_hash = "abcdef123456789"
    model_1_revision_1.refs = {"main", "refs/pr/1"}
    model_1_revision_1.last_modified = 123456789000  # 2 years ago
    model_1_revision_1.last_modified_str = "2 years ago"

    model_1.revisions = {model_1_revision_1}

    # Second model with 2 revisions
    model_2 = Mock()
    model_2.repo_type = "model"
    model_2.repo_id = "dummy_model"
    model_2.size_on_disk_str = "1.4K"
    model_2.last_accessed = 1550000000
    model_2.last_accessed_str = "2 years ago"
    model_2.size_on_disk = 1.4 * 1024  # 1.4K

    model_2_revision_1 = Mock()
    model_2_revision_1.commit_hash = "recent_hash_id"
    model_2_revision_1.refs = {"main"}
    model_2_revision_1.last_modified = 123456789  # 2 years ago
    model_2_revision_1.last_modified_str = "2 years ago"

    model_2_revision_2 = Mock()
    model_2_revision_2.commit_hash = "older_hash_id"
    model_2_revision_2.refs = {}
    model_2_revision_2.last_modified = 12345678000  # 3 years ago
    model_2_revision_2.last_modified_str = "3 years ago"

    model_2.revisions = {model_2_revision_1, model_2_revision_2}

    # And a dataset with 1 revision
    dataset_1 = Mock()
    dataset_1.repo_type = "dataset"
    dataset_1.repo_id = "dummy_dataset"
    dataset_1.size_on_disk_str = "8M"
    dataset_1.last_accessed = 1659000000
    dataset_1.last_accessed_str = "2 weeks ago"
    dataset_1.size_on_disk = 8 * 1024**2  # 8 MiB

    dataset_1_revision_1 = Mock()
    dataset_1_revision_1.commit_hash = "dataset_revision_hash_id"
    dataset_1_revision_1.refs = {}
    dataset_1_revision_1.last_modified = 1234567890000  # 1 day ago (newest)
    dataset_1_revision_1.last_modified_str = "1 day ago"

    dataset_1.revisions = {dataset_1_revision_1}

    # Fake cache
    strategy_mock = Mock()
    strategy_mock.repos = []
    strategy_mock.snapshots = []
    strategy_mock.expected_freed_size_str = "5.1M"

    cache_mock = Mock()
    cache_mock.repos = {model_1, model_2, dataset_1}
    cache_mock.delete_revisions.return_value = strategy_mock
    return cache_mock



from io import BytesIO

import pytest
import requests

from huggingface_hub.file_download import http_get

from .testing_utils import (
    OfflineSimulationMode,
    RequestWouldHangIndefinitelyError,
    offline,
)


def test_offline_with_timeout():
    with offline(OfflineSimulationMode.CONNECTION_TIMES_OUT):
        with pytest.raises(RequestWouldHangIndefinitelyError):
            requests.request("GET", "https://huggingface.co")
        with pytest.raises(requests.exceptions.ConnectTimeout):
            requests.request("GET", "https://huggingface.co", timeout=1.0)
        with pytest.raises(requests.exceptions.ConnectTimeout):
            http_get("https://huggingface.co", BytesIO())


def test_offline_with_connection_error():
    with offline(OfflineSimulationMode.CONNECTION_FAILS):
        with pytest.raises(requests.exceptions.ConnectionError):
            requests.request("GET", "https://huggingface.co")
        with pytest.raises(requests.exceptions.ConnectionError):
            http_get("https://huggingface.co", BytesIO())


def test_offline_with_datasets_offline_mode_enabled():
    with offline(OfflineSimulationMode.HF_HUB_OFFLINE_SET_TO_1):
        with pytest.raises(ConnectionError):
            http_get("https://huggingface.co", BytesIO())



# Copyright 2025 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains commands to interact with jobs on the Hugging Face Hub.

Usage:
    # run a job
    hf jobs run <image> <command>

    # List running or completed jobs
    hf jobs ps [-a] [-f key=value] [--format TEMPLATE]

    # Stream logs from a job
    hf jobs logs <job-id>

    # Inspect detailed information about a job
    hf jobs inspect <job-id>

    # Cancel a running job
    hf jobs cancel <job-id>
"""

import json
import os
import re
from argparse import Namespace, _SubParsersAction
from dataclasses import asdict
from pathlib import Path
from typing import Dict, List, Optional, Union

import requests

from huggingface_hub import HfApi, SpaceHardware, get_token
from huggingface_hub.utils import logging
from huggingface_hub.utils._dotenv import load_dotenv

from . import BaseHuggingfaceCLICommand


logger = logging.get_logger(__name__)

SUGGESTED_FLAVORS = [item.value for item in SpaceHardware if item.value != "zero-a10g"]


class JobsCommands(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction):
        jobs_parser = parser.add_parser("jobs", help="Run and manage Jobs on the Hub.")
        jobs_subparsers = jobs_parser.add_subparsers(help="huggingface.co jobs related commands")

        # Show help if no subcommand is provided
        jobs_parser.set_defaults(func=lambda args: jobs_parser.print_help())

        # Register commands
        InspectCommand.register_subcommand(jobs_subparsers)
        LogsCommand.register_subcommand(jobs_subparsers)
        PsCommand.register_subcommand(jobs_subparsers)
        RunCommand.register_subcommand(jobs_subparsers)
        CancelCommand.register_subcommand(jobs_subparsers)
        UvCommand.register_subcommand(jobs_subparsers)
        ScheduledJobsCommands.register_subcommand(jobs_subparsers)


class RunCommand(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction) -> None:
        run_parser = parser.add_parser("run", help="Run a Job")
        run_parser.add_argument("image", type=str, help="The Docker image to use.")
        run_parser.add_argument("-e", "--env", action="append", help="Set environment variables. E.g. --env ENV=value")
        run_parser.add_argument(
            "-s",
            "--secrets",
            action="append",
            help=(
                "Set secret environment variables. E.g. --secrets SECRET=value "
                "or `--secrets HF_TOKEN` to pass your Hugging Face token."
            ),
        )
        run_parser.add_argument("--env-file", type=str, help="Read in a file of environment variables.")
        run_parser.add_argument("--secrets-file", type=str, help="Read in a file of secret environment variables.")
        run_parser.add_argument(
            "--flavor",
            type=str,
            help=f"Flavor for the hardware, as in HF Spaces. Defaults to `cpu-basic`. Possible values: {', '.join(SUGGESTED_FLAVORS)}.",
        )
        run_parser.add_argument(
            "--timeout",
            type=str,
            help="Max duration: int/float with s (seconds, default), m (minutes), h (hours) or d (days).",
        )
        run_parser.add_argument(
            "-d",
            "--detach",
            action="store_true",
            help="Run the Job in the background and print the Job ID.",
        )
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace where the Job will be created. Defaults to the current user's namespace.",
        )
        run_parser.add_argument(
            "--token",
            type=str,
            help="A User Access Token generated from https://huggingface.co/settings/tokens",
        )
        run_parser.add_argument("command", nargs="...", help="The command to run.")
        run_parser.set_defaults(func=RunCommand)

    def __init__(self, args: Namespace) -> None:
        self.image: str = args.image
        self.command: List[str] = args.command
        self.env: dict[str, Optional[str]] = {}
        if args.env_file:
            self.env.update(load_dotenv(Path(args.env_file).read_text(), environ=os.environ.copy()))
        for env_value in args.env or []:
            self.env.update(load_dotenv(env_value, environ=os.environ.copy()))
        self.secrets: dict[str, Optional[str]] = {}
        extended_environ = _get_extended_environ()
        if args.secrets_file:
            self.secrets.update(load_dotenv(Path(args.secrets_file).read_text(), environ=extended_environ))
        for secret in args.secrets or []:
            self.secrets.update(load_dotenv(secret, environ=extended_environ))
        self.flavor: Optional[SpaceHardware] = args.flavor
        self.timeout: Optional[str] = args.timeout
        self.detach: bool = args.detach
        self.namespace: Optional[str] = args.namespace
        self.token: Optional[str] = args.token

    def run(self) -> None:
        api = HfApi(token=self.token)
        job = api.run_job(
            image=self.image,
            command=self.command,
            env=self.env,
            secrets=self.secrets,
            flavor=self.flavor,
            timeout=self.timeout,
            namespace=self.namespace,
        )
        # Always print the job ID to the user
        print(f"Job started with ID: {job.id}")
        print(f"View at: {job.url}")

        if self.detach:
            return

        # Now let's stream the logs
        for log in api.fetch_job_logs(job_id=job.id):
            print(log)


class LogsCommand(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction) -> None:
        run_parser = parser.add_parser("logs", help="Fetch the logs of a Job")
        run_parser.add_argument("job_id", type=str, help="Job ID")
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace where the job is running. Defaults to the current user's namespace.",
        )
        run_parser.add_argument(
            "--token", type=str, help="A User Access Token generated from https://huggingface.co/settings/tokens"
        )
        run_parser.set_defaults(func=LogsCommand)

    def __init__(self, args: Namespace) -> None:
        self.job_id: str = args.job_id
        self.namespace: Optional[str] = args.namespace
        self.token: Optional[str] = args.token

    def run(self) -> None:
        api = HfApi(token=self.token)
        for log in api.fetch_job_logs(job_id=self.job_id, namespace=self.namespace):
            print(log)


def _tabulate(rows: List[List[Union[str, int]]], headers: List[str]) -> str:
    """
    Inspired by:

    - stackoverflow.com/a/8356620/593036
    - stackoverflow.com/questions/9535954/printing-lists-as-tabular-data
    """
    col_widths = [max(len(str(x)) for x in col) for col in zip(*rows, headers)]
    terminal_width = max(os.get_terminal_size().columns, len(headers) * 12)
    while len(headers) + sum(col_widths) > terminal_width:
        col_to_minimize = col_widths.index(max(col_widths))
        col_widths[col_to_minimize] //= 2
        if len(headers) + sum(col_widths) <= terminal_width:
            col_widths[col_to_minimize] = terminal_width - sum(col_widths) - len(headers) + col_widths[col_to_minimize]
    row_format = ("{{:{}}} " * len(headers)).format(*col_widths)
    lines = []
    lines.append(row_format.format(*headers))
    lines.append(row_format.format(*["-" * w for w in col_widths]))
    for row in rows:
        row_format_args = [
            str(x)[: col_width - 3] + "..." if len(str(x)) > col_width else str(x)
            for x, col_width in zip(row, col_widths)
        ]
        lines.append(row_format.format(*row_format_args))
    return "\n".join(lines)


class PsCommand(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction) -> None:
        run_parser = parser.add_parser("ps", help="List Jobs")
        run_parser.add_argument(
            "-a",
            "--all",
            action="store_true",
            help="Show all Jobs (default shows just running)",
        )
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace from where it lists the jobs. Defaults to the current user's namespace.",
        )
        run_parser.add_argument(
            "--token",
            type=str,
            help="A User Access Token generated from https://huggingface.co/settings/tokens",
        )
        # Add Docker-style filtering argument
        run_parser.add_argument(
            "-f",
            "--filter",
            action="append",
            default=[],
            help="Filter output based on conditions provided (format: key=value)",
        )
        # Add option to format output
        run_parser.add_argument(
            "--format",
            type=str,
            help="Format output using a custom template",
        )
        run_parser.set_defaults(func=PsCommand)

    def __init__(self, args: Namespace) -> None:
        self.all: bool = args.all
        self.namespace: Optional[str] = args.namespace
        self.token: Optional[str] = args.token
        self.format: Optional[str] = args.format
        self.filters: Dict[str, str] = {}

        # Parse filter arguments (key=value pairs)
        for f in args.filter:
            if "=" in f:
                key, value = f.split("=", 1)
                self.filters[key.lower()] = value
            else:
                print(f"Warning: Ignoring invalid filter format '{f}'. Use key=value format.")

    def run(self) -> None:
        """
        Fetch and display job information for the current user.
        Uses Docker-style filtering with -f/--filter flag and key=value pairs.
        """
        try:
            api = HfApi(token=self.token)

            # Fetch jobs data
            jobs = api.list_jobs(namespace=self.namespace)

            # Define table headers
            table_headers = ["JOB ID", "IMAGE/SPACE", "COMMAND", "CREATED", "STATUS"]

            # Process jobs data
            rows = []

            for job in jobs:
                # Extract job data for filtering
                status = job.status.stage if job.status else "UNKNOWN"

                # Skip job if not all jobs should be shown and status doesn't match criteria
                if not self.all and status not in ("RUNNING", "UPDATING"):
                    continue

                # Extract job ID
                job_id = job.id

                # Extract image or space information
                image_or_space = job.docker_image or "N/A"

                # Extract and format command
                command = job.command or []
                command_str = " ".join(command) if command else "N/A"

                # Extract creation time
                created_at = job.created_at.strftime("%Y-%m-%d %H:%M:%S") if job.created_at else "N/A"

                # Create a dict with all job properties for filtering
                job_properties = {
                    "id": job_id,
                    "image": image_or_space,
                    "status": status.lower(),
                    "command": command_str,
                }

                # Check if job matches all filters
                if not self._matches_filters(job_properties):
                    continue

                # Create row
                rows.append([job_id, image_or_space, command_str, created_at, status])

            # Handle empty results
            if not rows:
                filters_msg = ""
                if self.filters:
                    filters_msg = f" matching filters: {', '.join([f'{k}={v}' for k, v in self.filters.items()])}"

                print(f"No jobs found{filters_msg}")
                return

            # Apply custom format if provided or use default tabular format
            self._print_output(rows, table_headers)

        except requests.RequestException as e:
            print(f"Error fetching jobs data: {e}")
        except (KeyError, ValueError, TypeError) as e:
            print(f"Error processing jobs data: {e}")
        except Exception as e:
            print(f"Unexpected error - {type(e).__name__}: {e}")

    def _matches_filters(self, job_properties: Dict[str, str]) -> bool:
        """Check if job matches all specified filters."""
        for key, pattern in self.filters.items():
            # Check if property exists
            if key not in job_properties:
                return False

            # Support pattern matching with wildcards
            if "*" in pattern or "?" in pattern:
                # Convert glob pattern to regex
                regex_pattern = pattern.replace("*", ".*").replace("?", ".")
                if not re.search(f"^{regex_pattern}$", job_properties[key], re.IGNORECASE):
                    return False
            # Simple substring matching
            elif pattern.lower() not in job_properties[key].lower():
                return False

        return True

    def _print_output(self, rows, headers):
        """Print output according to the chosen format."""
        if self.format:
            # Custom template formatting (simplified)
            template = self.format
            for row in rows:
                line = template
                for i, field in enumerate(["id", "image", "command", "created", "status"]):
                    placeholder = f"{{{{.{field}}}}}"
                    if placeholder in line:
                        line = line.replace(placeholder, str(row[i]))
                print(line)
        else:
            # Default tabular format
            print(
                _tabulate(
                    rows,
                    headers=headers,
                )
            )


class InspectCommand(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction) -> None:
        run_parser = parser.add_parser("inspect", help="Display detailed information on one or more Jobs")
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace where the job is running. Defaults to the current user's namespace.",
        )
        run_parser.add_argument(
            "--token", type=str, help="A User Access Token generated from https://huggingface.co/settings/tokens"
        )
        run_parser.add_argument("job_ids", nargs="...", help="The jobs to inspect")
        run_parser.set_defaults(func=InspectCommand)

    def __init__(self, args: Namespace) -> None:
        self.namespace: Optional[str] = args.namespace
        self.token: Optional[str] = args.token
        self.job_ids: List[str] = args.job_ids

    def run(self) -> None:
        api = HfApi(token=self.token)
        jobs = [api.inspect_job(job_id=job_id, namespace=self.namespace) for job_id in self.job_ids]
        print(json.dumps([asdict(job) for job in jobs], indent=4, default=str))


class CancelCommand(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction) -> None:
        run_parser = parser.add_parser("cancel", help="Cancel a Job")
        run_parser.add_argument("job_id", type=str, help="Job ID")
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace where the job is running. Defaults to the current user's namespace.",
        )
        run_parser.add_argument(
            "--token", type=str, help="A User Access Token generated from https://huggingface.co/settings/tokens"
        )
        run_parser.set_defaults(func=CancelCommand)

    def __init__(self, args: Namespace) -> None:
        self.job_id: str = args.job_id
        self.namespace = args.namespace
        self.token: Optional[str] = args.token

    def run(self) -> None:
        api = HfApi(token=self.token)
        api.cancel_job(job_id=self.job_id, namespace=self.namespace)


class UvCommand(BaseHuggingfaceCLICommand):
    """Run UV scripts on Hugging Face infrastructure."""

    @staticmethod
    def register_subcommand(parser):
        """Register UV run subcommand."""
        uv_parser = parser.add_parser(
            "uv",
            help="Run UV scripts (Python with inline dependencies) on HF infrastructure",
        )

        subparsers = uv_parser.add_subparsers(dest="uv_command", help="UV commands", required=True)

        # Run command only
        run_parser = subparsers.add_parser(
            "run",
            help="Run a UV script (local file or URL) on HF infrastructure",
        )
        run_parser.add_argument("script", help="UV script to run (local file or URL)")
        run_parser.add_argument("script_args", nargs="...", help="Arguments for the script", default=[])
        run_parser.add_argument("--image", type=str, help="Use a custom Docker image with `uv` installed.")
        run_parser.add_argument(
            "--repo",
            help="Repository name for the script (creates ephemeral if not specified)",
        )
        run_parser.add_argument(
            "--flavor",
            type=str,
            help=f"Flavor for the hardware, as in HF Spaces. Defaults to `cpu-basic`. Possible values: {', '.join(SUGGESTED_FLAVORS)}.",
        )
        run_parser.add_argument("-e", "--env", action="append", help="Environment variables")
        run_parser.add_argument(
            "-s",
            "--secrets",
            action="append",
            help=(
                "Set secret environment variables. E.g. --secrets SECRET=value "
                "or `--secrets HF_TOKEN` to pass your Hugging Face token."
            ),
        )
        run_parser.add_argument("--env-file", type=str, help="Read in a file of environment variables.")
        run_parser.add_argument(
            "--secrets-file",
            type=str,
            help="Read in a file of secret environment variables.",
        )
        run_parser.add_argument("--timeout", type=str, help="Max duration (e.g., 30s, 5m, 1h)")
        run_parser.add_argument("-d", "--detach", action="store_true", help="Run in background")
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace where the Job will be created. Defaults to the current user's namespace.",
        )
        run_parser.add_argument("--token", type=str, help="HF token")
        # UV options
        run_parser.add_argument("--with", action="append", help="Run with the given packages installed", dest="with_")
        run_parser.add_argument(
            "-p", "--python", type=str, help="The Python interpreter to use for the run environment"
        )
        run_parser.set_defaults(func=UvCommand)

    def __init__(self, args: Namespace) -> None:
        """Initialize the command with parsed arguments."""
        self.script = args.script
        self.script_args = args.script_args
        self.dependencies = args.with_
        self.python = args.python
        self.image = args.image
        self.env: dict[str, Optional[str]] = {}
        if args.env_file:
            self.env.update(load_dotenv(Path(args.env_file).read_text(), environ=os.environ.copy()))
        for env_value in args.env or []:
            self.env.update(load_dotenv(env_value, environ=os.environ.copy()))
        self.secrets: dict[str, Optional[str]] = {}
        extended_environ = _get_extended_environ()
        if args.secrets_file:
            self.secrets.update(load_dotenv(Path(args.secrets_file).read_text(), environ=extended_environ))
        for secret in args.secrets or []:
            self.secrets.update(load_dotenv(secret, environ=extended_environ))
        self.flavor: Optional[SpaceHardware] = args.flavor
        self.timeout: Optional[str] = args.timeout
        self.detach: bool = args.detach
        self.namespace: Optional[str] = args.namespace
        self.token: Optional[str] = args.token
        self._repo = args.repo

    def run(self) -> None:
        """Execute UV command."""
        logging.set_verbosity(logging.INFO)
        api = HfApi(token=self.token)
        job = api.run_uv_job(
            script=self.script,
            script_args=self.script_args,
            dependencies=self.dependencies,
            python=self.python,
            image=self.image,
            env=self.env,
            secrets=self.secrets,
            flavor=self.flavor,
            timeout=self.timeout,
            namespace=self.namespace,
            _repo=self._repo,
        )

        # Always print the job ID to the user
        print(f"Job started with ID: {job.id}")
        print(f"View at: {job.url}")

        if self.detach:
            return

        # Now let's stream the logs
        for log in api.fetch_job_logs(job_id=job.id):
            print(log)


def _get_extended_environ() -> Dict[str, str]:
    extended_environ = os.environ.copy()
    if (token := get_token()) is not None:
        extended_environ["HF_TOKEN"] = token
    return extended_environ


class ScheduledJobsCommands(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction):
        scheduled_jobs_parser = parser.add_parser("scheduled", help="Create and manage scheduled Jobs on the Hub.")
        scheduled_jobs_subparsers = scheduled_jobs_parser.add_subparsers(
            help="huggingface.co scheduled jobs related commands"
        )

        # Show help if no subcommand is provided
        scheduled_jobs_parser.set_defaults(func=lambda args: scheduled_jobs_subparsers.print_help())

        # Register commands
        ScheduledRunCommand.register_subcommand(scheduled_jobs_subparsers)
        ScheduledPsCommand.register_subcommand(scheduled_jobs_subparsers)
        ScheduledInspectCommand.register_subcommand(scheduled_jobs_subparsers)
        ScheduledDeleteCommand.register_subcommand(scheduled_jobs_subparsers)
        ScheduledSuspendCommand.register_subcommand(scheduled_jobs_subparsers)
        ScheduledResumeCommand.register_subcommand(scheduled_jobs_subparsers)
        ScheduledUvCommand.register_subcommand(scheduled_jobs_subparsers)


class ScheduledRunCommand(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction) -> None:
        run_parser = parser.add_parser("run", help="Schedule a Job")
        run_parser.add_argument(
            "schedule",
            type=str,
            help="One of annually, yearly, monthly, weekly, daily, hourly, or a CRON schedule expression.",
        )
        run_parser.add_argument("image", type=str, help="The Docker image to use.")
        run_parser.add_argument(
            "--suspend",
            action="store_true",
            help="Suspend (pause) the scheduled Job",
            default=None,
        )
        run_parser.add_argument(
            "--concurrency",
            action="store_true",
            help="Allow multiple instances of this Job to run concurrently",
            default=None,
        )
        run_parser.add_argument("-e", "--env", action="append", help="Set environment variables. E.g. --env ENV=value")
        run_parser.add_argument(
            "-s",
            "--secrets",
            action="append",
            help=(
                "Set secret environment variables. E.g. --secrets SECRET=value "
                "or `--secrets HF_TOKEN` to pass your Hugging Face token."
            ),
        )
        run_parser.add_argument("--env-file", type=str, help="Read in a file of environment variables.")
        run_parser.add_argument("--secrets-file", type=str, help="Read in a file of secret environment variables.")
        run_parser.add_argument(
            "--flavor",
            type=str,
            help=f"Flavor for the hardware, as in HF Spaces. Defaults to `cpu-basic`. Possible values: {', '.join(SUGGESTED_FLAVORS)}.",
        )
        run_parser.add_argument(
            "--timeout",
            type=str,
            help="Max duration: int/float with s (seconds, default), m (minutes), h (hours) or d (days).",
        )
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace where the scheduled Job will be created. Defaults to the current user's namespace.",
        )
        run_parser.add_argument(
            "--token",
            type=str,
            help="A User Access Token generated from https://huggingface.co/settings/tokens",
        )
        run_parser.add_argument("command", nargs="...", help="The command to run.")
        run_parser.set_defaults(func=ScheduledRunCommand)

    def __init__(self, args: Namespace) -> None:
        self.schedule: str = args.schedule
        self.image: str = args.image
        self.command: List[str] = args.command
        self.suspend: Optional[bool] = args.suspend
        self.concurrency: Optional[bool] = args.concurrency
        self.env: dict[str, Optional[str]] = {}
        if args.env_file:
            self.env.update(load_dotenv(Path(args.env_file).read_text(), environ=os.environ.copy()))
        for env_value in args.env or []:
            self.env.update(load_dotenv(env_value, environ=os.environ.copy()))
        self.secrets: dict[str, Optional[str]] = {}
        extended_environ = _get_extended_environ()
        if args.secrets_file:
            self.secrets.update(load_dotenv(Path(args.secrets_file).read_text(), environ=extended_environ))
        for secret in args.secrets or []:
            self.secrets.update(load_dotenv(secret, environ=extended_environ))
        self.flavor: Optional[SpaceHardware] = args.flavor
        self.timeout: Optional[str] = args.timeout
        self.namespace: Optional[str] = args.namespace
        self.token: Optional[str] = args.token

    def run(self) -> None:
        api = HfApi(token=self.token)
        scheduled_job = api.create_scheduled_job(
            image=self.image,
            command=self.command,
            schedule=self.schedule,
            suspend=self.suspend,
            concurrency=self.concurrency,
            env=self.env,
            secrets=self.secrets,
            flavor=self.flavor,
            timeout=self.timeout,
            namespace=self.namespace,
        )
        # Always print the scheduled job ID to the user
        print(f"Scheduled Job created with ID: {scheduled_job.id}")


class ScheduledPsCommand(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction) -> None:
        run_parser = parser.add_parser("ps", help="List scheduled Jobs")
        run_parser.add_argument(
            "-a",
            "--all",
            action="store_true",
            help="Show all scheduled Jobs (default hides suspended)",
        )
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace from where it lists the jobs. Defaults to the current user's namespace.",
        )
        run_parser.add_argument(
            "--token",
            type=str,
            help="A User Access Token generated from https://huggingface.co/settings/tokens",
        )
        # Add Docker-style filtering argument
        run_parser.add_argument(
            "-f",
            "--filter",
            action="append",
            default=[],
            help="Filter output based on conditions provided (format: key=value)",
        )
        # Add option to format output
        run_parser.add_argument(
            "--format",
            type=str,
            help="Format output using a custom template",
        )
        run_parser.set_defaults(func=ScheduledPsCommand)

    def __init__(self, args: Namespace) -> None:
        self.all: bool = args.all
        self.namespace: Optional[str] = args.namespace
        self.token: Optional[str] = args.token
        self.format: Optional[str] = args.format
        self.filters: Dict[str, str] = {}

        # Parse filter arguments (key=value pairs)
        for f in args.filter:
            if "=" in f:
                key, value = f.split("=", 1)
                self.filters[key.lower()] = value
            else:
                print(f"Warning: Ignoring invalid filter format '{f}'. Use key=value format.")

    def run(self) -> None:
        """
        Fetch and display scheduked job information for the current user.
        Uses Docker-style filtering with -f/--filter flag and key=value pairs.
        """
        try:
            api = HfApi(token=self.token)

            # Fetch jobs data
            scheduled_jobs = api.list_scheduled_jobs(namespace=self.namespace)

            # Define table headers
            table_headers = [
                "ID",
                "SCHEDULE",
                "IMAGE/SPACE",
                "COMMAND",
                "LAST RUN",
                "NEXT RUN",
                "SUSPEND",
            ]

            # Process jobs data
            rows = []

            for scheduled_job in scheduled_jobs:
                # Extract job data for filtering
                suspend = scheduled_job.suspend

                # Skip job if not all jobs should be shown and status doesn't match criteria
                if not self.all and suspend:
                    continue

                # Extract job ID
                scheduled_job_id = scheduled_job.id

                # Extract schedule
                schedule = scheduled_job.schedule

                # Extract image or space information
                image_or_space = scheduled_job.job_spec.docker_image or "N/A"

                # Extract and format command
                command = scheduled_job.job_spec.command or []
                command_str = " ".join(command) if command else "N/A"

                # Extract status
                last_job_at = (
                    scheduled_job.status.last_job.at.strftime("%Y-%m-%d %H:%M:%S")
                    if scheduled_job.status.last_job
                    else "N/A"
                )
                next_job_run_at = (
                    scheduled_job.status.next_job_run_at.strftime("%Y-%m-%d %H:%M:%S")
                    if scheduled_job.status.next_job_run_at
                    else "N/A"
                )

                # Create a dict with all job properties for filtering
                job_properties = {
                    "id": scheduled_job_id,
                    "image": image_or_space,
                    "suspend": str(suspend),
                    "command": command_str,
                }

                # Check if job matches all filters
                if not self._matches_filters(job_properties):
                    continue

                # Create row
                rows.append(
                    [
                        scheduled_job_id,
                        schedule,
                        image_or_space,
                        command_str,
                        last_job_at,
                        next_job_run_at,
                        suspend,
                    ]
                )

            # Handle empty results
            if not rows:
                filters_msg = ""
                if self.filters:
                    filters_msg = f" matching filters: {', '.join([f'{k}={v}' for k, v in self.filters.items()])}"

                print(f"No scheduled jobs found{filters_msg}")
                return

            # Apply custom format if provided or use default tabular format
            self._print_output(rows, table_headers)

        except requests.RequestException as e:
            print(f"Error fetching scheduled jobs data: {e}")
        except (KeyError, ValueError, TypeError) as e:
            print(f"Error processing scheduled jobs data: {e}")
        except Exception as e:
            print(f"Unexpected error - {type(e).__name__}: {e}")

    def _matches_filters(self, job_properties: Dict[str, str]) -> bool:
        """Check if scheduled job matches all specified filters."""
        for key, pattern in self.filters.items():
            # Check if property exists
            if key not in job_properties:
                return False

            # Support pattern matching with wildcards
            if "*" in pattern or "?" in pattern:
                # Convert glob pattern to regex
                regex_pattern = pattern.replace("*", ".*").replace("?", ".")
                if not re.search(f"^{regex_pattern}$", job_properties[key], re.IGNORECASE):
                    return False
            # Simple substring matching
            elif pattern.lower() not in job_properties[key].lower():
                return False

        return True

    def _print_output(self, rows, headers):
        """Print output according to the chosen format."""
        if self.format:
            # Custom template formatting (simplified)
            template = self.format
            for row in rows:
                line = template
                for i, field in enumerate(
                    ["id", "schedule", "image", "command", "last_job_at", "next_job_run_at", "suspend"]
                ):
                    placeholder = f"{{{{.{field}}}}}"
                    if placeholder in line:
                        line = line.replace(placeholder, str(row[i]))
                print(line)
        else:
            # Default tabular format
            print(
                _tabulate(
                    rows,
                    headers=headers,
                )
            )


class ScheduledInspectCommand(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction) -> None:
        run_parser = parser.add_parser("inspect", help="Display detailed information on one or more scheduled Jobs")
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace where the scheduled job is. Defaults to the current user's namespace.",
        )
        run_parser.add_argument(
            "--token", type=str, help="A User Access Token generated from https://huggingface.co/settings/tokens"
        )
        run_parser.add_argument("scheduled_job_ids", nargs="...", help="The scheduled jobs to inspect")
        run_parser.set_defaults(func=ScheduledInspectCommand)

    def __init__(self, args: Namespace) -> None:
        self.namespace: Optional[str] = args.namespace
        self.token: Optional[str] = args.token
        self.scheduled_job_ids: List[str] = args.scheduled_job_ids

    def run(self) -> None:
        api = HfApi(token=self.token)
        scheduled_jobs = [
            api.inspect_scheduled_job(scheduled_job_id=scheduled_job_id, namespace=self.namespace)
            for scheduled_job_id in self.scheduled_job_ids
        ]
        print(json.dumps([asdict(scheduled_job) for scheduled_job in scheduled_jobs], indent=4, default=str))


class ScheduledDeleteCommand(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction) -> None:
        run_parser = parser.add_parser("delete", help="Delete a scheduled Job")
        run_parser.add_argument("scheduled_job_id", type=str, help="Scheduled Job ID")
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace where the scheduled job is. Defaults to the current user's namespace.",
        )
        run_parser.add_argument(
            "--token", type=str, help="A User Access Token generated from https://huggingface.co/settings/tokens"
        )
        run_parser.set_defaults(func=ScheduledDeleteCommand)

    def __init__(self, args: Namespace) -> None:
        self.scheduled_job_id: str = args.scheduled_job_id
        self.namespace = args.namespace
        self.token: Optional[str] = args.token

    def run(self) -> None:
        api = HfApi(token=self.token)
        api.delete_scheduled_job(scheduled_job_id=self.scheduled_job_id, namespace=self.namespace)


class ScheduledSuspendCommand(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction) -> None:
        run_parser = parser.add_parser("suspend", help="Suspend (pause) a scheduled Job")
        run_parser.add_argument("scheduled_job_id", type=str, help="Scheduled Job ID")
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace where the scheduled job is. Defaults to the current user's namespace.",
        )
        run_parser.add_argument(
            "--token", type=str, help="A User Access Token generated from https://huggingface.co/settings/tokens"
        )
        run_parser.set_defaults(func=ScheduledSuspendCommand)

    def __init__(self, args: Namespace) -> None:
        self.scheduled_job_id: str = args.scheduled_job_id
        self.namespace = args.namespace
        self.token: Optional[str] = args.token

    def run(self) -> None:
        api = HfApi(token=self.token)
        api.suspend_scheduled_job(scheduled_job_id=self.scheduled_job_id, namespace=self.namespace)


class ScheduledResumeCommand(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction) -> None:
        run_parser = parser.add_parser("resume", help="Resume (unpause) a scheduled Job")
        run_parser.add_argument("scheduled_job_id", type=str, help="Scheduled Job ID")
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace where the scheduled job is. Defaults to the current user's namespace.",
        )
        run_parser.add_argument(
            "--token", type=str, help="A User Access Token generated from https://huggingface.co/settings/tokens"
        )
        run_parser.set_defaults(func=ScheduledResumeCommand)

    def __init__(self, args: Namespace) -> None:
        self.scheduled_job_id: str = args.scheduled_job_id
        self.namespace = args.namespace
        self.token: Optional[str] = args.token

    def run(self) -> None:
        api = HfApi(token=self.token)
        api.resume_scheduled_job(scheduled_job_id=self.scheduled_job_id, namespace=self.namespace)


class ScheduledUvCommand(BaseHuggingfaceCLICommand):
    """Schedule UV scripts on Hugging Face infrastructure."""

    @staticmethod
    def register_subcommand(parser):
        """Register UV run subcommand."""
        uv_parser = parser.add_parser(
            "uv",
            help="Schedule UV scripts (Python with inline dependencies) on HF infrastructure",
        )

        subparsers = uv_parser.add_subparsers(dest="uv_command", help="UV commands", required=True)

        # Run command only
        run_parser = subparsers.add_parser(
            "run",
            help="Run a UV script (local file or URL) on HF infrastructure",
        )
        run_parser.add_argument(
            "schedule",
            type=str,
            help="One of annually, yearly, monthly, weekly, daily, hourly, or a CRON schedule expression.",
        )
        run_parser.add_argument("script", help="UV script to run (local file or URL)")
        run_parser.add_argument("script_args", nargs="...", help="Arguments for the script", default=[])
        run_parser.add_argument(
            "--suspend",
            action="store_true",
            help="Suspend (pause) the scheduled Job",
            default=None,
        )
        run_parser.add_argument(
            "--concurrency",
            action="store_true",
            help="Allow multiple instances of this Job to run concurrently",
            default=None,
        )
        run_parser.add_argument("--image", type=str, help="Use a custom Docker image with `uv` installed.")
        run_parser.add_argument(
            "--repo",
            help="Repository name for the script (creates ephemeral if not specified)",
        )
        run_parser.add_argument(
            "--flavor",
            type=str,
            help=f"Flavor for the hardware, as in HF Spaces. Defaults to `cpu-basic`. Possible values: {', '.join(SUGGESTED_FLAVORS)}.",
        )
        run_parser.add_argument("-e", "--env", action="append", help="Environment variables")
        run_parser.add_argument(
            "-s",
            "--secrets",
            action="append",
            help=(
                "Set secret environment variables. E.g. --secrets SECRET=value "
                "or `--secrets HF_TOKEN` to pass your Hugging Face token."
            ),
        )
        run_parser.add_argument("--env-file", type=str, help="Read in a file of environment variables.")
        run_parser.add_argument(
            "--secrets-file",
            type=str,
            help="Read in a file of secret environment variables.",
        )
        run_parser.add_argument("--timeout", type=str, help="Max duration (e.g., 30s, 5m, 1h)")
        run_parser.add_argument("-d", "--detach", action="store_true", help="Run in background")
        run_parser.add_argument(
            "--namespace",
            type=str,
            help="The namespace where the Job will be created. Defaults to the current user's namespace.",
        )
        run_parser.add_argument("--token", type=str, help="HF token")
        # UV options
        run_parser.add_argument("--with", action="append", help="Run with the given packages installed", dest="with_")
        run_parser.add_argument(
            "-p", "--python", type=str, help="The Python interpreter to use for the run environment"
        )
        run_parser.set_defaults(func=ScheduledUvCommand)

    def __init__(self, args: Namespace) -> None:
        """Initialize the command with parsed arguments."""
        self.schedule: str = args.schedule
        self.script = args.script
        self.script_args = args.script_args
        self.suspend: Optional[bool] = args.suspend
        self.concurrency: Optional[bool] = args.concurrency
        self.dependencies = args.with_
        self.python = args.python
        self.image = args.image
        self.env: dict[str, Optional[str]] = {}
        if args.env_file:
            self.env.update(load_dotenv(Path(args.env_file).read_text(), environ=os.environ.copy()))
        for env_value in args.env or []:
            self.env.update(load_dotenv(env_value, environ=os.environ.copy()))
        self.secrets: dict[str, Optional[str]] = {}
        extended_environ = _get_extended_environ()
        if args.secrets_file:
            self.secrets.update(load_dotenv(Path(args.secrets_file).read_text(), environ=extended_environ))
        for secret in args.secrets or []:
            self.secrets.update(load_dotenv(secret, environ=extended_environ))
        self.flavor: Optional[SpaceHardware] = args.flavor
        self.timeout: Optional[str] = args.timeout
        self.detach: bool = args.detach
        self.namespace: Optional[str] = args.namespace
        self.token: Optional[str] = args.token
        self._repo = args.repo

    def run(self) -> None:
        """Schedule UV command."""
        logging.set_verbosity(logging.INFO)
        api = HfApi(token=self.token)
        job = api.create_scheduled_uv_job(
            script=self.script,
            script_args=self.script_args,
            schedule=self.schedule,
            suspend=self.suspend,
            concurrency=self.concurrency,
            dependencies=self.dependencies,
            python=self.python,
            image=self.image,
            env=self.env,
            secrets=self.secrets,
            flavor=self.flavor,
            timeout=self.timeout,
            namespace=self.namespace,
            _repo=self._repo,
        )

        # Always print the job ID to the user
        print(f"Scheduled Job created with ID: {job.id}")



import inspect
from dataclasses import asdict, astuple, dataclass, is_dataclass
from typing import Any, Dict, List, Literal, Optional, Set, Tuple, Union, get_type_hints

import jedi
import pytest

from huggingface_hub.dataclasses import _is_validator, as_validated_field, strict, type_validator, validated_field
from huggingface_hub.errors import (
    StrictDataclassClassValidationError,
    StrictDataclassDefinitionError,
    StrictDataclassFieldValidationError,
)


def positive_int(value: int):
    if not value >= 0:
        raise ValueError(f"Value must be positive, got {value}")


def multiple_of_64(value: int):
    if not value % 64 == 0:
        raise ValueError(f"Value must be a multiple of 64, got {value}")


@as_validated_field
def strictly_positive(value: int):
    if not value > 0:
        raise ValueError(f"Value must be strictly positive, got {value}")


def dtype_validation(value: "ForwardDtype"):
    if not isinstance(value, str):
        raise ValueError(f"Value must be string, got {value}")

    if isinstance(value, str) and value not in ["float32", "bfloat16", "float16"]:
        raise ValueError(f"Value must be one of `[float32, bfloat16, float16]` but got {value}")


@strict
@dataclass
class ConfigForwardRef:
    """Test forward reference handling.

    In practice, forward reference types are not validated so a custom validator is highly recommended.
    """

    forward_ref_validated: "ForwardDtype" = validated_field(validator=dtype_validation)
    forward_ref: "ForwardDtype" = "float32"  # type is not validated by default


class ForwardDtype(str):
    """Dummy class to simulate a forward reference (e.g. `torch.dtype`)."""


@strict
@dataclass
class Config:
    model_type: str
    hidden_size: int = validated_field(validator=[positive_int, multiple_of_64])
    vocab_size: int = strictly_positive(default=16)


@strict(accept_kwargs=True)
@dataclass
class ConfigWithKwargs:
    model_type: str
    vocab_size: int = validated_field(validator=positive_int, default=16)


class DummyClass:
    pass


def test_valid_initialization():
    config = Config(model_type="bert", vocab_size=30000, hidden_size=768)
    assert config.model_type == "bert"
    assert config.vocab_size == 30000
    assert config.hidden_size == 768


def test_default_values():
    config = Config(model_type="bert", hidden_size=1024)
    assert config.model_type == "bert"
    assert config.vocab_size == 16
    assert config.hidden_size == 1024


def test_forward_ref_validation_is_skipped():
    config = ConfigForwardRef(forward_ref="float32", forward_ref_validated="float32")
    assert config.forward_ref == "float32"
    assert config.forward_ref_validated == "float32"

    # The `forward_ref_validated` has proper validation added in field-metadata and will be validated
    with pytest.raises(StrictDataclassFieldValidationError):
        ConfigForwardRef(forward_ref_validated="float64")

    with pytest.raises(StrictDataclassFieldValidationError):
        ConfigForwardRef(forward_ref_validated=-1)

    with pytest.raises(StrictDataclassFieldValidationError):
        ConfigForwardRef(forward_ref_validated="not_dtype")

    # The `forward_ref` type is not validated => user can input anything
    ConfigForwardRef(forward_ref=-1, forward_ref_validated="float32")
    ConfigForwardRef(forward_ref=["float32"], forward_ref_validated="float32")


def test_invalid_type_initialization():
    with pytest.raises(StrictDataclassFieldValidationError):
        Config(model_type={"type": "bert"}, vocab_size=30000, hidden_size=768)

    with pytest.raises(StrictDataclassFieldValidationError):
        Config(model_type="bert", vocab_size="30000", hidden_size=768)


def test_all_validators_are_applied():
    # must be positive
    with pytest.raises(StrictDataclassFieldValidationError):
        Config(model_type="bert", vocab_size=-1, hidden_size=1024)

    # must be a multiple of 64
    with pytest.raises(StrictDataclassFieldValidationError):
        Config(model_type="bert", hidden_size=1025)

    # both validators are applied
    with pytest.raises(StrictDataclassFieldValidationError):
        Config(model_type="bert", hidden_size=-1024)


def test_validated_on_assignment():
    config = Config(model_type="bert", hidden_size=1024)
    config.vocab_size = 10000  # ok
    with pytest.raises(StrictDataclassFieldValidationError):
        config.vocab_size = "10000"  # type validator checked
    with pytest.raises(StrictDataclassFieldValidationError):
        config.vocab_size = -1  # custom validators checked
    with pytest.raises(StrictDataclassFieldValidationError):
        config.vocab_size = 0  # must be strictly positive


def test_lax_on_new_attributes():
    config = Config(model_type="bert", hidden_size=1024)
    config.new_attribute = "new_value"
    assert config.new_attribute == "new_value"  # not validated


def test_custom_validator_must_be_callable():
    """Must raise at class definition time."""
    with pytest.raises(StrictDataclassDefinitionError):

        @strict
        @dataclass
        class Config:
            model_type: str = validated_field(validator="not_a_function")

    with pytest.raises(StrictDataclassDefinitionError):

        @strict
        @dataclass
        class Config:
            model_type: str = validated_field(validator=lambda: None)  # not a validator either


@pytest.mark.parametrize(
    "value, type_annotation",
    [
        # Basic types
        (5, int),
        (5.0, float),
        ("John", str),
        # Union types
        (5, Union[int, str]),
        ("John", Union[int, str]),
        # Optional
        (5, Optional[int]),
        (None, Optional[int]),
        (DummyClass(), Optional[DummyClass]),
        # Literal
        ("John", Literal["John", "Doe"]),
        (5, Literal[4, 5, 6]),
        # List
        ([1, 2, 3], List[int]),
        ([1, 2, "3"], List[Union[int, str]]),
        # Tuple
        ((1, 2, 3), Tuple[int, int, int]),
        ((1, 2, "3"), Tuple[int, int, str]),
        ((1, 2, 3, 4), Tuple[int, ...]),
        # Dict
        ({"a": 1, "b": 2}, Dict[str, int]),
        ({"a": 1, "b": "2"}, Dict[str, Union[int, str]]),
        # Set
        ({1, 2, 3}, Set[int]),
        ({1, 2, "3"}, Set[Union[int, str]]),
        # Custom classes
        (DummyClass(), DummyClass),
        # Any
        (5, Any),
        ("John", Any),
        (DummyClass(), Any),
        # Deep nested type
        (
            {
                "a": [
                    (1, DummyClass(), {1, "2", "3", 4}),
                    (2, DummyClass(), None),
                ],
            },
            Dict[
                str,
                List[
                    Tuple[
                        int,
                        DummyClass,
                        Optional[Set[Union[int, str],]],
                    ]
                ],
            ],
        ),
    ],
)
def test_type_validator_valid(value, type_annotation):
    type_validator("dummy", value, type_annotation)


@pytest.mark.parametrize(
    "value, type_annotation",
    [
        # Basic types
        (5, float),
        (5.0, int),
        ("John", int),
        # Union types
        (5.0, Union[int, str]),
        (None, Union[int, str]),
        (DummyClass(), Union[int, str]),
        # Optional
        ("John", Optional[int]),
        (DummyClass(), Optional[int]),
        # Literal
        ("Ada", Literal["John", "Doe"]),
        (3, Literal[4, 5, 6]),
        # List
        (5, List[int]),
        ([1, 2, "3"], List[int]),
        # Tuple
        (5, Tuple[int, int, int]),
        ((1, 2, "3"), Tuple[int, int, int]),
        ((1, 2, 3, 4), Tuple[int, int, int]),
        ((1, 2, "3", 4), Tuple[int, ...]),
        # Dict
        (5, Dict[str, int]),
        ({"a": 1, "b": "2"}, Dict[str, int]),
        # Set
        (5, Set[int]),
        ({1, 2, "3"}, Set[int]),
        # Custom classes
        (5, DummyClass),
        ("John", DummyClass),
    ],
)
def test_type_validator_invalid(value, type_annotation):
    with pytest.raises(TypeError):
        type_validator("dummy", value, type_annotation)


class DummyValidator:
    def __init__(self, threshold):
        self.threshold = threshold

    def __call__(self, value):
        return value < self.threshold

    def compare(self, value, value2=10):
        return value < value2


@pytest.mark.parametrize(
    "obj",
    [
        positive_int,
        multiple_of_64,
        lambda value: None,
        lambda value, factor=2: None,
        lambda value=1, factor=2: value * factor,
        lambda *values: None,
        DummyValidator(threshold=10),  # callable object
        DummyValidator(threshold=10).compare,  # callable method
    ],
)
def test_is_validator(obj):
    # Anything that can be called with `obj(value)` is a correct validator.
    assert _is_validator(obj)


@pytest.mark.parametrize(
    "obj",
    [
        5,  # not callable
        lambda: None,  # no argument
        lambda value1, value2: None,  # more than one argument with default values
        lambda *, value: None,  # keyword-only argument
    ],
)
def test_not_a_validator(obj):
    assert not _is_validator(obj)


def test_preserve_metadata():
    class ConfigWithMetadataField:
        foo: int = strictly_positive(metadata={"foo": "bar"}, default=10)

    assert ConfigWithMetadataField.foo.metadata["foo"] == "bar"


def test_accept_kwargs():
    config = ConfigWithKwargs(model_type="bert", vocab_size=30000, hidden_size=768)
    assert config.model_type == "bert"
    assert config.vocab_size == 30000
    assert config.hidden_size == 768

    # Defined fields are still validated
    with pytest.raises(StrictDataclassFieldValidationError):
        ConfigWithKwargs(model_type="bert", vocab_size=-1)

    # Default values are still used
    config = ConfigWithKwargs(model_type="bert")
    assert config.vocab_size == 16


def test_do_not_accept_kwargs():
    @strict
    @dataclass
    class Config:
        model_type: str

    with pytest.raises(TypeError):
        Config(model_type="bert", vocab_size=30000)


def test_is_recognized_as_dataclass():
    # Check that dataclasses module recognizes it as a dataclass
    assert is_dataclass(Config)

    # Check that an instance is recognized as a dataclass instance
    config = Config(model_type="bert", hidden_size=768)
    assert is_dataclass(config)


def test_behave_as_a_dataclass():
    # Check that dataclasses.asdict works
    config = Config(model_type="bert", hidden_size=768)
    assert asdict(config) == {"model_type": "bert", "hidden_size": 768, "vocab_size": 16}

    # Check that dataclasses.astuple works
    assert astuple(config) == ("bert", 768, 16)


def test_type_annotations_preserved():
    # Check that type hints are preserved
    hints = get_type_hints(Config)
    assert hints["model_type"] is str
    assert hints["hidden_size"] is int
    assert hints["vocab_size"] is int


def test_correct_init_signature():
    # Check that __init__ has the expected signature
    signature = inspect.signature(Config.__init__)
    parameters = list(signature.parameters.values())

    # First param should be self
    assert parameters[0].name == "self"

    # model_type should be required
    assert parameters[1].name == "model_type"
    assert parameters[1].default == inspect.Parameter.empty

    # hidden_size should be required (and validated)
    assert parameters[2].name == "hidden_size"
    assert parameters[2].default == inspect.Parameter.empty

    # vocab_size should be optional with default
    assert parameters[3].name == "vocab_size"
    assert parameters[3].default == 16


def test_correct_eq_repr():
    # Test equality comparison
    config1 = Config(model_type="bert", hidden_size=0)
    config2 = Config(model_type="bert", hidden_size=0)
    config3 = Config(model_type="gpt", hidden_size=0)

    assert config1 == config2
    assert config1 != config3

    # Test repr
    assert repr(config1) == "Config(model_type='bert', hidden_size=0, vocab_size=16)"


def test_repr_if_accept_kwargs():
    config1 = ConfigWithKwargs(foo="bar", model_type="bert")
    assert repr(config1) == "ConfigWithKwargs(model_type='bert', vocab_size=16, *foo='bar')"


def test_autocompletion_attribute_without_kwargs():
    # Create a sample script
    completions = jedi.Script("""
from dataclasses import dataclass
from huggingface_hub.dataclasses import strict

@strict
@dataclass
class Config:
    model_type: str
    hidden_size: int = 768

config = Config(model_type="bert")
config.
""").complete(line=12, column=7)
    completion_names = [c.name for c in completions]
    assert "model_type" in completion_names
    assert "hidden_size" in completion_names


def test_autocompletion_attribute_with_kwargs():
    # Create a sample script
    completions = jedi.Script("""
from dataclasses import dataclass
from huggingface_hub.dataclasses import strict

@strict(accept_kwargs=True)
@dataclass
class Config:
    model_type: str
    hidden_size: int = 768

config = Config(model_type="bert", foo="bar")
config.
""").complete(line=12, column=7)
    completion_names = [c.name for c in completions]
    assert "model_type" in completion_names
    assert "hidden_size" in completion_names
    assert "foo" not in completion_names  # not an official arg


def test_autocompletion_init_without_kwargs():
    # Create a sample script
    completions = jedi.Script("""
from dataclasses import dataclass
from huggingface_hub.dataclasses import strict

@strict
@dataclass
class Config:
    model_type: str
    hidden_size: int = 768

config = Config(
""").complete(line=11, column=16)
    completion_names = [c.name for c in completions]
    assert "model_type=" in completion_names
    assert "hidden_size=" in completion_names


def test_autocompletion_init_with_kwargs():
    # Create a sample script
    completions = jedi.Script("""
from dataclasses import dataclass
from huggingface_hub.dataclasses import strict

@strict(accept_kwargs=True)
@dataclass
class Config:
    model_type: str
    hidden_size: int = 768

config = Config(
""").complete(line=11, column=16)
    completion_names = [c.name for c in completions]
    assert "model_type=" in completion_names
    assert "hidden_size=" in completion_names


def test_strict_requires_dataclass():
    with pytest.raises(StrictDataclassDefinitionError):

        @strict
        class InvalidConfig:
            model_type: str


class TestClassValidation:
    @strict
    @dataclass
    class ParentConfig:
        foo: str = "bar"
        foo_length: int = 3

        def validate_foo_length(self):
            if len(self.foo) != self.foo_length:
                raise ValueError(f"foo must be {self.foo_length} characters long, got {len(self.foo)}")

    @strict
    @dataclass
    class ChildConfig(ParentConfig):
        number: int = 42

        def validate_number_multiple_of_foo_length(self):
            if self.number % self.foo_length != 0:
                raise ValueError(f"number must be a multiple of foo_length ({self.foo_length}), got {self.number}")

    @strict
    @dataclass
    class OtherChildConfig(ParentConfig):
        number: int = 42

    @strict
    @dataclass
    class ChildConfigWithPostInit(ParentConfig):
        def __post_init__(self):
            # Let's assume post_init doubles each value
            # Validation is ran AFTER __post_init__
            self.foo = self.foo * 2
            self.foo_length = self.foo_length * 2

    def test_parent_config_validation(self):
        # Test valid initialization
        config = self.ParentConfig(foo="bar", foo_length=3)
        assert config.foo == "bar"
        assert config.foo_length == 3

        # Test invalid initialization
        with pytest.raises(StrictDataclassClassValidationError):
            self.ParentConfig(foo="bar", foo_length=4)

    def test_child_config_validation(self):
        # Test valid initialization
        config = self.ChildConfig(foo="bar", foo_length=3, number=42)
        assert config.foo == "bar"
        assert config.foo_length == 3
        assert config.number == 42

        # Test invalid initialization
        with pytest.raises(StrictDataclassClassValidationError):
            self.ChildConfig(foo="bar", foo_length=4, number=40)

        with pytest.raises(StrictDataclassClassValidationError):
            self.ChildConfig(foo="bar", foo_length=3, number=43)

    def test_other_child_config_validation(self):
        # Test valid initialization
        config = self.OtherChildConfig(foo="bar", foo_length=3, number=43)
        assert config.foo == "bar"
        assert config.foo_length == 3
        assert config.number == 43  # not validated => did not fail

        # Test invalid initialization
        with pytest.raises(StrictDataclassClassValidationError):
            self.OtherChildConfig(foo="bar", foo_length=4, number=42)

    def test_validate_after_init(self):
        # Test valid initialization
        config = self.ParentConfig(foo="bar", foo_length=3)

        # Attributes can be updated after initialization
        config.foo = "abcd"
        config.foo_length = 4
        config.validate()  # Explicit call required

        # Explicit validation fails
        config.foo_length = 5
        with pytest.raises(StrictDataclassClassValidationError):
            config.validate()

    def test_validation_runs_after_post_init(self):
        config = self.ChildConfigWithPostInit(foo="bar", foo_length=3)
        assert config.foo == "barbar"
        assert config.foo_length == 6

        with pytest.raises(StrictDataclassClassValidationError, match="foo must be 4 characters long, got 6"):
            # post init doubles the value and then the validation fails
            self.ChildConfigWithPostInit(foo="bar", foo_length=2)


class TestClassValidationWithInheritance:
    """Regression test.

    If parent class is not a strict dataclass but defines validators, the child class should validate them too.
    """

    class Base:
        def validate_foo(self):
            if self.foo < 0:
                raise ValueError("foo must be positive")

    @strict
    @dataclass
    class Config(Base):
        foo: int
        bar: int

        def validate_bar(self):
            if self.bar < 0:
                raise ValueError("bar must be positive")

    def test_class_validation_with_inheritance(self):
        # Test valid initialization
        config = self.Config(foo=0, bar=0)
        assert config.foo == 0
        assert config.bar == 0

        # Test invalid initialization
        with pytest.raises(StrictDataclassClassValidationError):
            self.Config(foo=0, bar=-1)  # validation from child class

        with pytest.raises(StrictDataclassClassValidationError):
            self.Config(foo=-1, bar=0)  # validation from parent class


class TestClassValidateAlreadyExists:
    """Regression test.

    If a class already has a validate method, it should raise a StrictDataclassDefinitionError.
    """

    def test_validate_already_defined_by_class(self):
        with pytest.raises(StrictDataclassDefinitionError):

            @strict
            @dataclass
            class Config:
                foo: int = 0

                def validate(self):
                    pass  # already defined => should raise an error

    def test_validate_already_defined_by_parent(self):
        with pytest.raises(StrictDataclassDefinitionError):

            class ParentClass:
                def validate(self):
                    pass

            @strict
            @dataclass
            class ConfigWithParent(ParentClass):  # 'validate' already defined => should raise an error
                foo: int = 0



# Copyright 2020 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from pathlib import Path
from unittest.mock import patch

import pytest
from PIL import Image

from huggingface_hub import hf_hub_download
from huggingface_hub.inference_api import InferenceApi

from .testing_utils import expect_deprecation, with_production_testing


@pytest.mark.vcr
@with_production_testing
class InferenceApiTest(unittest.TestCase):
    def read(self, filename: str) -> bytes:
        return Path(filename).read_bytes()

    @classmethod
    @with_production_testing
    def setUpClass(cls) -> None:
        cls.image_file = hf_hub_download(repo_id="Narsil/image_dummy", repo_type="dataset", filename="lena.png")
        return super().setUpClass()

    @expect_deprecation("huggingface_hub.inference_api")
    def test_simple_inference(self):
        api = InferenceApi("bert-base-uncased")
        inputs = "Hi, I think [MASK] is cool"
        results = api(inputs)
        self.assertIsInstance(results, list)

        result = results[0]
        self.assertIsInstance(result, dict)
        self.assertTrue("sequence" in result)
        self.assertTrue("score" in result)

    @unittest.skip("Model often not loaded")
    @expect_deprecation("huggingface_hub.inference_api")
    def test_inference_with_params(self):
        api = InferenceApi("typeform/distilbert-base-uncased-mnli")
        inputs = "I bought a device but it is not working and I would like to get reimbursed!"
        params = {"candidate_labels": ["refund", "legal", "faq"]}
        result = api(inputs, params)
        self.assertIsInstance(result, dict)
        self.assertTrue("sequence" in result)
        self.assertTrue("scores" in result)

    @unittest.skip("Model often not loaded")
    @expect_deprecation("huggingface_hub.inference_api")
    def test_inference_with_dict_inputs(self):
        api = InferenceApi("distilbert-base-cased-distilled-squad")
        inputs = {
            "question": "What's my name?",
            "context": "My name is Clara and I live in Berkeley.",
        }
        result = api(inputs)
        self.assertIsInstance(result, dict)
        self.assertTrue("score" in result)
        self.assertTrue("answer" in result)

    @unittest.skip("Model often not loaded")
    @expect_deprecation("huggingface_hub.inference_api")
    def test_inference_with_audio(self):
        api = InferenceApi("facebook/wav2vec2-base-960h")
        file = hf_hub_download(
            repo_id="hf-internal-testing/dummy-flac-single-example",
            repo_type="dataset",
            filename="example.flac",
        )
        data = self.read(file)
        result = api(data=data)
        self.assertIsInstance(result, dict)
        self.assertTrue("text" in result, f"We received {result} instead")

    @unittest.skip("Model often not loaded")
    @expect_deprecation("huggingface_hub.inference_api")
    def test_inference_with_image(self):
        api = InferenceApi("google/vit-base-patch16-224")
        data = self.read(self.image_file)
        result = api(data=data)
        self.assertIsInstance(result, list)
        for classification in result:
            self.assertIsInstance(classification, dict)
            self.assertTrue("score" in classification)
            self.assertTrue("label" in classification)

    @expect_deprecation("huggingface_hub.inference_api")
    def test_text_to_image(self):
        api = InferenceApi("stabilityai/stable-diffusion-2-1")
        with patch("huggingface_hub.inference_api.get_session") as mock:
            mock().post.return_value.headers = {"Content-Type": "image/jpeg"}
            mock().post.return_value.content = self.read(self.image_file)
            output = api("cat")
        self.assertIsInstance(output, Image.Image)

    @expect_deprecation("huggingface_hub.inference_api")
    def test_text_to_image_raw_response(self):
        api = InferenceApi("stabilityai/stable-diffusion-2-1")
        with patch("huggingface_hub.inference_api.get_session") as mock:
            mock().post.return_value.headers = {"Content-Type": "image/jpeg"}
            mock().post.return_value.content = self.read(self.image_file)
            output = api("cat", raw_response=True)
        # Raw response is returned
        self.assertEqual(output, mock().post.return_value)

    @expect_deprecation("huggingface_hub.inference_api")
    def test_inference_overriding_task(self):
        api = InferenceApi(
            "sentence-transformers/paraphrase-albert-small-v2",
            task="feature-extraction",
        )
        inputs = "This is an example again"
        result = api(inputs)
        self.assertIsInstance(result, list)

    @expect_deprecation("huggingface_hub.inference_api")
    def test_inference_overriding_invalid_task(self):
        with self.assertRaises(ValueError, msg="Invalid task invalid-task. Make sure it's valid."):
            InferenceApi("bert-base-uncased", task="invalid-task")

    @expect_deprecation("huggingface_hub.inference_api")
    def test_inference_missing_input(self):
        api = InferenceApi("deepset/roberta-base-squad2")
        result = api({"question": "What's my name?"})
        self.assertIsInstance(result, dict)
        self.assertTrue("error" in result)



import sys
import unittest

from huggingface_hub.utils import is_tf_available


def require_tf(test_case):
    """
    Decorator marking a test that requires TensorFlow.

    These tests are skipped when TensorFlow is not installed.

    """
    if not is_tf_available():
        return unittest.skip("test requires Tensorflow")(test_case)
    else:
        return test_case


@require_tf
def test_import_huggingface_hub_does_not_import_tensorflow():
    # `import huggingface_hub` is not necessary since huggingface_hub is already imported at the top of this file,
    # but let's keep it here anyway just in case
    import huggingface_hub  # noqa

    assert "tensorflow" not in sys.modules



import json
import os
import unittest
from pathlib import Path

import pytest

from huggingface_hub import HfApi, hf_hub_download, snapshot_download
from huggingface_hub.keras_mixin import (
    KerasModelHubMixin,
    from_pretrained_keras,
    push_to_hub_keras,
    save_pretrained_keras,
)
from huggingface_hub.utils import is_graphviz_available, is_pydot_available, is_tf_available, logging

from .testing_constants import ENDPOINT_STAGING, TOKEN, USER
from .testing_utils import repo_name


logger = logging.get_logger(__name__)


if is_tf_available():
    import tensorflow as tf


def require_tf(test_case):
    """
    Decorator marking a test that requires TensorFlow, graphviz and pydot.

    These tests are skipped when TensorFlow, graphviz and pydot are installed.

    """
    if not is_tf_available() or not is_pydot_available() or not is_graphviz_available():
        return unittest.skip("test requires Tensorflow, graphviz and pydot.")(test_case)
    else:
        return test_case


if is_tf_available():
    # Define dummy mixin model...
    class DummyModel(tf.keras.Model, KerasModelHubMixin):
        def __init__(self, **kwargs):
            super().__init__()
            self.l1 = tf.keras.layers.Dense(2, activation="relu")
            dummy_batch_size = input_dim = 2
            self.dummy_inputs = tf.ones([dummy_batch_size, input_dim])

        def call(self, x):
            return self.l1(x)

else:
    DummyModel = None


@require_tf
@pytest.mark.usefixtures("fx_cache_dir")
class CommonKerasTest(unittest.TestCase):
    cache_dir: Path

    @classmethod
    def setUpClass(cls):
        """
        Share this valid token in all tests below.
        """
        cls._api = HfApi(endpoint=ENDPOINT_STAGING, token=TOKEN)


class HubMixinTestKeras(CommonKerasTest):
    def test_save_pretrained(self):
        model = DummyModel()
        model(model.dummy_inputs)

        model.save_pretrained(self.cache_dir)
        files = os.listdir(self.cache_dir)
        self.assertTrue("saved_model.pb" in files)
        self.assertTrue("keras_metadata.pb" in files)
        self.assertTrue("README.md" in files)
        self.assertTrue("model.png" in files)
        self.assertEqual(len(files), 7)

        model.save_pretrained(self.cache_dir, config={"num": 12, "act": "gelu"})
        files = os.listdir(self.cache_dir)
        self.assertTrue("config.json" in files)
        self.assertTrue("saved_model.pb" in files)
        self.assertEqual(len(files), 8)

    def test_keras_from_pretrained_weights(self):
        model = DummyModel()
        model(model.dummy_inputs)

        model.save_pretrained(self.cache_dir)
        new_model = DummyModel.from_pretrained(self.cache_dir)

        # Check the reloaded model's weights match the original model's weights
        self.assertTrue(tf.reduce_all(tf.equal(new_model.weights[0], model.weights[0])))

        # Check a new model's weights are not the same as the reloaded model's weights
        another_model = DummyModel()
        another_model(tf.ones([2, 2]))
        self.assertFalse(tf.reduce_all(tf.equal(new_model.weights[0], another_model.weights[0])).numpy().item())

    def test_abs_path_from_pretrained(self):
        model = DummyModel()
        model(model.dummy_inputs)
        model.save_pretrained(self.cache_dir, config={"num": 10, "act": "gelu_fast"})
        model = DummyModel.from_pretrained(self.cache_dir)
        self.assertTrue(model.config == {"num": 10, "act": "gelu_fast"})

    def test_push_to_hub_keras_mixin_via_http_basic(self):
        repo_id = f"{USER}/{repo_name()}"

        model = DummyModel()
        model(model.dummy_inputs)

        model.push_to_hub(repo_id=repo_id, token=TOKEN, config={"num": 7, "act": "gelu_fast"})

        # Test model id exists
        assert self._api.model_info(repo_id).id == repo_id

        # Test config has been pushed to hub
        config_path = hf_hub_download(
            repo_id=repo_id, filename="config.json", use_auth_token=TOKEN, cache_dir=self.cache_dir
        )
        with open(config_path) as f:
            assert json.load(f) == {"num": 7, "act": "gelu_fast"}

        # Delete tmp file and repo
        self._api.delete_repo(repo_id=repo_id)


@require_tf
class HubKerasSequentialTest(CommonKerasTest):
    def model_init(self):
        model = tf.keras.models.Sequential()
        model.add(tf.keras.layers.Dense(2, activation="relu"))
        model.compile(optimizer="adam", loss="mse")
        return model

    def model_fit(self, model):
        x = tf.constant([[0.44, 0.90], [0.65, 0.39]])
        y = tf.constant([[1, 1], [0, 0]])
        model.fit(x, y)
        return model

    def test_save_pretrained(self):
        model = self.model_init()
        with pytest.raises(ValueError, match="Model should be built*"):
            save_pretrained_keras(model, save_directory=self.cache_dir)
        model.build((None, 2))

        save_pretrained_keras(model, save_directory=self.cache_dir)
        files = os.listdir(self.cache_dir)
        self.assertIn("saved_model.pb", files)
        self.assertIn("keras_metadata.pb", files)
        self.assertIn("model.png", files)
        self.assertIn("README.md", files)
        self.assertEqual(len(files), 7)

        loaded_model = from_pretrained_keras(self.cache_dir)
        self.assertIsNone(loaded_model.optimizer)

    def test_save_pretrained_model_card_fit(self):
        model = self.model_init()
        model = self.model_fit(model)

        save_pretrained_keras(model, save_directory=self.cache_dir)
        files = os.listdir(self.cache_dir)
        history = json.loads((self.cache_dir / "history.json").read_text())

        self.assertIn("saved_model.pb", files)
        self.assertIn("keras_metadata.pb", files)
        self.assertIn("model.png", files)
        self.assertIn("README.md", files)
        self.assertIn("history.json", files)
        self.assertEqual(history, model.history.history)
        self.assertEqual(len(files), 8)

    def test_save_model_card_history_removal(self):
        model = self.model_init()
        model = self.model_fit(model)

        history_path = self.cache_dir / "history.json"
        history_path.write_text("Keras FTW")

        with pytest.warns(UserWarning, match="`history.json` file already exists, *"):
            save_pretrained_keras(model, save_directory=self.cache_dir)
            # assert that it's not the same as old history file and it's overridden
            self.assertNotEqual("Keras FTW", history_path.read_text())

            # Check the history is saved as a json in the repository.
            files = os.listdir(self.cache_dir)
            self.assertIn("history.json", files)

            # Check that there is no "Training Metrics" section in the model card.
            # This was done in an older version.
            self.assertNotIn("Training Metrics", (self.cache_dir / "README.md").read_text())

    def test_save_pretrained_optimizer_state(self):
        model = self.model_init()
        model.build((None, 2))
        save_pretrained_keras(model, self.cache_dir, include_optimizer=True)
        loaded_model = from_pretrained_keras(self.cache_dir)
        self.assertIsNotNone(loaded_model.optimizer)

    def test_from_pretrained_weights(self):
        model = self.model_init()
        model.build((None, 2))

        save_pretrained_keras(model, self.cache_dir)
        new_model = from_pretrained_keras(self.cache_dir)

        # Check a new model's weights are not the same as the reloaded model's weights
        another_model = DummyModel()
        another_model(tf.ones([2, 2]))
        self.assertFalse(tf.reduce_all(tf.equal(new_model.weights[0], another_model.weights[0])).numpy().item())

    def test_save_pretrained_task_name_deprecation(self):
        model = self.model_init()
        model.build((None, 2))

        with pytest.warns(
            FutureWarning,
            match="`task_name` input argument is deprecated. Pass `tags` instead.",
        ):
            save_pretrained_keras(model, self.cache_dir, tags=["test"], task_name="test", save_traces=True)

    def test_abs_path_from_pretrained(self):
        model = self.model_init()
        model.build((None, 2))
        save_pretrained_keras(
            model, self.cache_dir, config={"num": 10, "act": "gelu_fast"}, plot_model=True, tags=None
        )
        new_model = from_pretrained_keras(self.cache_dir)
        self.assertTrue(tf.reduce_all(tf.equal(new_model.weights[0], model.weights[0])))
        self.assertTrue(new_model.config == {"num": 10, "act": "gelu_fast"})

    def test_push_to_hub_keras_sequential_via_http_basic(self):
        repo_id = f"{USER}/{repo_name()}"
        model = self.model_init()
        model = self.model_fit(model)

        push_to_hub_keras(model, repo_id=repo_id, token=TOKEN, api_endpoint=ENDPOINT_STAGING)
        assert self._api.model_info(repo_id).id == repo_id
        repo_files = self._api.list_repo_files(repo_id)
        assert "README.md" in repo_files
        assert "model.png" in repo_files
        self._api.delete_repo(repo_id=repo_id)

    def test_push_to_hub_keras_sequential_via_http_plot_false(self):
        repo_id = f"{USER}/{repo_name()}"
        model = self.model_init()
        model = self.model_fit(model)

        push_to_hub_keras(model, repo_id=repo_id, token=TOKEN, api_endpoint=ENDPOINT_STAGING, plot_model=False)
        repo_files = self._api.list_repo_files(repo_id)
        self.assertNotIn("model.png", repo_files)
        self._api.delete_repo(repo_id=repo_id)

    def test_push_to_hub_keras_via_http_override_tensorboard(self):
        """Test log directory is overwritten when pushing a keras model a 2nd time."""
        repo_id = f"{USER}/{repo_name()}"

        log_dir = self.cache_dir / "tb_log_dir"
        log_dir.mkdir(parents=True, exist_ok=True)
        (log_dir / "tensorboard.txt").write_text("Keras FTW")

        model = self.model_init()
        model.build((None, 2))
        push_to_hub_keras(model, repo_id=repo_id, log_dir=log_dir, api_endpoint=ENDPOINT_STAGING, token=TOKEN)

        log_dir2 = self.cache_dir / "tb_log_dir2"
        log_dir2.mkdir(parents=True, exist_ok=True)
        (log_dir2 / "override.txt").write_text("Keras FTW")
        push_to_hub_keras(model, repo_id=repo_id, log_dir=log_dir2, api_endpoint=ENDPOINT_STAGING, token=TOKEN)

        files = self._api.list_repo_files(repo_id)
        self.assertIn("logs/override.txt", files)
        self.assertNotIn("logs/tensorboard.txt", files)

        self._api.delete_repo(repo_id=repo_id)

    def test_push_to_hub_keras_via_http_with_model_kwargs(self):
        repo_id = f"{USER}/{repo_name()}"

        model = self.model_init()
        model = self.model_fit(model)
        push_to_hub_keras(
            model,
            repo_id=repo_id,
            api_endpoint=ENDPOINT_STAGING,
            token=TOKEN,
            include_optimizer=True,
            save_traces=False,
        )

        assert self._api.model_info(repo_id).id == repo_id

        snapshot_path = snapshot_download(repo_id=repo_id, cache_dir=self.cache_dir)
        from_pretrained_keras(snapshot_path)

        self._api.delete_repo(repo_id)


@require_tf
class HubKerasFunctionalTest(CommonKerasTest):
    def model_init(self):
        inputs = tf.keras.layers.Input(shape=(2,))
        outputs = tf.keras.layers.Dense(2, activation="relu")(inputs)
        model = tf.keras.models.Model(inputs=inputs, outputs=outputs)
        model.compile(optimizer="adam", loss="mse")
        return model

    def model_fit(self, model):
        x = tf.constant([[0.44, 0.90], [0.65, 0.39]])
        y = tf.constant([[1, 1], [0, 0]])
        model.fit(x, y)
        return model

    def test_save_pretrained(self):
        model = self.model_init()
        model.build((None, 2))
        self.assertTrue(model.built)

        save_pretrained_keras(model, self.cache_dir)
        files = os.listdir(self.cache_dir)

        self.assertIn("saved_model.pb", files)
        self.assertIn("keras_metadata.pb", files)
        self.assertEqual(len(files), 7)

    def test_save_pretrained_fit(self):
        model = self.model_init()
        model = self.model_fit(model)

        save_pretrained_keras(model, self.cache_dir)
        files = os.listdir(self.cache_dir)

        self.assertIn("saved_model.pb", files)
        self.assertIn("keras_metadata.pb", files)
        self.assertEqual(len(files), 8)



# Copyright 2020 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import time
import unittest
from pathlib import Path

import pytest
import requests

from huggingface_hub import RepoUrl
from huggingface_hub.hf_api import HfApi
from huggingface_hub.repository import (
    Repository,
    is_tracked_upstream,
    is_tracked_with_lfs,
)
from huggingface_hub.utils import SoftTemporaryDirectory, logging, run_subprocess

from .testing_constants import ENDPOINT_STAGING, TOKEN
from .testing_utils import (
    expect_deprecation,
    repo_name,
    use_tmp_repo,
    with_production_testing,
)


logger = logging.get_logger(__name__)


@pytest.mark.usefixtures("fx_cache_dir")
class RepositoryTestAbstract(unittest.TestCase):
    cache_dir: Path
    repo_path: Path

    # This content is 5MB (under 10MB)
    small_content = json.dumps([100] * int(1e6))

    # This content is 20MB (over 10MB)
    large_content = json.dumps([100] * int(4e6))

    # This content is binary (contains the null character)
    binary_content = "\x00\x00\x00\x00"

    _api = HfApi(endpoint=ENDPOINT_STAGING, token=TOKEN)

    @classmethod
    def setUp(self) -> None:
        self.repo_path = self.cache_dir / "working_dir"
        self.repo_path.mkdir()

    def _create_dummy_files(self):
        # Create dummy files
        # one is lfs-tracked, the other is not.
        small_file = self.repo_path / "dummy.txt"
        small_file.write_text(self.small_content)

        binary_file = self.repo_path / "model.bin"
        binary_file.write_text(self.binary_content)


class TestRepositoryShared(RepositoryTestAbstract):
    """Tests in this class shares a single repo on the Hub (common to all tests).

    These tests must not push data to it.
    """

    @classmethod
    def setUpClass(cls):
        """
        Share this valid token in all tests below.
        """
        super().setUpClass()
        cls.repo_url = cls._api.create_repo(repo_id=repo_name())
        cls.repo_id = cls.repo_url.repo_id
        cls._api.upload_file(
            path_or_fileobj=cls.binary_content.encode(),
            path_in_repo="random_file.txt",
            repo_id=cls.repo_id,
        )

    @classmethod
    def tearDownClass(cls):
        cls._api.delete_repo(repo_id=cls.repo_id)

    @expect_deprecation("Repository")
    def test_clone_from_repo_url(self):
        Repository(self.repo_path, clone_from=self.repo_url)

    @expect_deprecation("Repository")
    def test_clone_from_repo_id(self):
        Repository(self.repo_path, clone_from=self.repo_id)

    @expect_deprecation("Repository")
    def test_clone_from_repo_name_no_namespace_fails(self):
        with self.assertRaises(EnvironmentError):
            Repository(self.repo_path, clone_from=self.repo_id.split("/")[1], token=TOKEN)

    @expect_deprecation("Repository")
    def test_clone_from_not_hf_url(self):
        # Should not error out
        Repository(self.repo_path, clone_from="https://hf.co/hf-internal-testing/huggingface-hub-dummy-repository")

    @expect_deprecation("Repository")
    def test_clone_from_missing_repo(self):
        """If the repo does not exist an EnvironmentError is raised."""
        with self.assertRaises(EnvironmentError):
            Repository(self.repo_path, clone_from="missing_repo")

    @expect_deprecation("Repository")
    @with_production_testing
    def test_clone_from_prod_canonical_repo_id(self):
        Repository(self.repo_path, clone_from="bert-base-cased", skip_lfs_files=True)

    @expect_deprecation("Repository")
    @with_production_testing
    def test_clone_from_prod_canonical_repo_url(self):
        Repository(self.repo_path, clone_from="https://huggingface.co/bert-base-cased", skip_lfs_files=True)

    @expect_deprecation("Repository")
    def test_init_from_existing_local_clone(self):
        run_subprocess(["git", "clone", self.repo_url, str(self.repo_path)])

        repo = Repository(self.repo_path)
        repo.lfs_track(["*.pdf"])
        repo.lfs_enable_largefiles()
        repo.git_pull()

    @expect_deprecation("Repository")
    def test_init_failure(self):
        with self.assertRaises(ValueError):
            Repository(self.repo_path)

    @expect_deprecation("Repository")
    def test_init_clone_in_empty_folder(self):
        repo = Repository(self.repo_path, clone_from=self.repo_url)
        repo.lfs_track(["*.pdf"])
        repo.lfs_enable_largefiles()
        repo.git_pull()
        self.assertIn("random_file.txt", os.listdir(self.repo_path))

    @expect_deprecation("Repository")
    def test_git_lfs_filename(self):
        run_subprocess("git init", folder=self.repo_path)

        repo = Repository(self.repo_path)
        large_file = self.repo_path / "large_file[].txt"
        large_file.write_text(self.large_content)

        repo.git_add()

        repo.lfs_track([large_file.name])
        self.assertFalse(is_tracked_with_lfs(large_file))

        repo.lfs_track([large_file.name], filename=True)
        self.assertTrue(is_tracked_with_lfs(large_file))

    @expect_deprecation("Repository")
    def test_init_clone_in_nonempty_folder(self):
        self._create_dummy_files()
        with self.assertRaises(EnvironmentError):
            Repository(self.repo_path, clone_from=self.repo_url)

    @expect_deprecation("Repository")
    def test_init_clone_in_nonempty_linked_git_repo_with_token(self):
        Repository(self.repo_path, clone_from=self.repo_url, token=TOKEN)
        Repository(self.repo_path, clone_from=self.repo_url, token=TOKEN)

    @expect_deprecation("Repository")
    def test_is_tracked_upstream(self):
        Repository(self.repo_path, clone_from=self.repo_id)
        self.assertTrue(is_tracked_upstream(self.repo_path))

    @expect_deprecation("Repository")
    def test_push_errors_on_wrong_checkout(self):
        repo = Repository(self.repo_path, clone_from=self.repo_id)

        head_commit_ref = run_subprocess("git show --oneline -s", folder=self.repo_path).stdout.split()[0]

        repo.git_checkout(head_commit_ref)

        with self.assertRaises(OSError):
            with repo.commit("New commit"):
                with open("new_file", "w+") as f:
                    f.write("Ok")


class TestRepositoryUniqueRepos(RepositoryTestAbstract):
    """Tests in this class use separated repos on the Hub (i.e. 1 test = 1 repo).

    These tests can push data to it.
    """

    def setUp(self):
        super().setUp()
        self.repo_url = self._api.create_repo(repo_id=repo_name())
        self.repo_id = self.repo_url.repo_id
        self._api.upload_file(
            path_or_fileobj=self.binary_content.encode(), path_in_repo="random_file.txt", repo_id=self.repo_id
        )

    def tearDown(self):
        self._api.delete_repo(repo_id=self.repo_id)

    @expect_deprecation("Repository")
    def clone_repo(self, **kwargs) -> Repository:
        if "local_dir" not in kwargs:
            kwargs["local_dir"] = self.repo_path
        if "clone_from" not in kwargs:
            kwargs["clone_from"] = self.repo_url
        if "token" not in kwargs:
            kwargs["token"] = TOKEN
        if "git_user" not in kwargs:
            kwargs["git_user"] = "ci"
        if "git_email" not in kwargs:
            kwargs["git_email"] = "ci@dummy.com"
        return Repository(**kwargs)

    @use_tmp_repo()
    @expect_deprecation("Repository")
    def test_init_clone_in_nonempty_non_linked_git_repo(self, repo_url: RepoUrl):
        self.clone_repo()

        # Try and clone another repository within the same directory.
        # Should error out due to mismatched remotes.
        with self.assertRaises(EnvironmentError):
            Repository(self.repo_path, clone_from=repo_url)

    def test_init_clone_in_nonempty_linked_git_repo(self):
        # Clone the repository to disk
        self.clone_repo()

        # Add to the remote repository without doing anything to the local repository.
        self._api.upload_file(
            path_or_fileobj=self.binary_content.encode(), path_in_repo="random_file_3.txt", repo_id=self.repo_id
        )

        # Cloning the repository in the same directory should not result in a git pull.
        self.clone_repo(clone_from=self.repo_url)
        self.assertNotIn("random_file_3.txt", os.listdir(self.repo_path))

    def test_init_clone_in_nonempty_linked_git_repo_unrelated_histories(self):
        # Clone the repository to disk
        repo = self.clone_repo()

        # Create and commit file locally
        (self.repo_path / "random_file_3.txt").write_text("hello world")
        repo.git_add()
        repo.git_commit("Unrelated commit")

        # Add to the remote repository without doing anything to the local repository.
        self._api.upload_file(
            path_or_fileobj=self.binary_content.encode(),
            path_in_repo="random_file_3.txt",
            repo_id=self.repo_url.repo_id,
        )

        # The repo should initialize correctly as the remote is the same, even with unrelated historied
        self.clone_repo()

    def test_add_commit_push(self):
        repo = self.clone_repo()
        self._create_dummy_files()
        repo.git_add()
        repo.git_commit()
        url = repo.git_push()

        # Check that the returned commit url
        # actually exists.
        r = requests.head(url)
        r.raise_for_status()

    def test_add_commit_push_non_blocking(self):
        repo = self.clone_repo()
        self._create_dummy_files()
        repo.git_add()
        repo.git_commit()
        url, result = repo.git_push(blocking=False)

        # Check background process
        if result._process.poll() is None:
            self.assertEqual(result.status, -1)

        while not result.is_done:
            time.sleep(0.5)

        self.assertTrue(result.is_done)
        self.assertEqual(result.status, 0)

        # Check that the returned commit url
        # actually exists.
        r = requests.head(url)
        r.raise_for_status()

    def test_context_manager_non_blocking(self):
        repo = self.clone_repo()

        with repo.commit("New commit", blocking=False):
            (self.repo_path / "dummy.txt").write_text("hello world")

        while repo.commands_in_progress:
            time.sleep(1)

        self.assertEqual(len(repo.commands_in_progress), 0)
        self.assertEqual(len(repo.command_queue), 1)
        self.assertEqual(repo.command_queue[-1].status, 0)
        self.assertEqual(repo.command_queue[-1].is_done, True)
        self.assertEqual(repo.command_queue[-1].title, "push")

    @unittest.skip("This is a flaky and legacy test")
    def test_add_commit_push_non_blocking_process_killed(self):
        repo = self.clone_repo()

        # Far too big file: will take forever
        (self.repo_path / "dummy.txt").write_text(str([[[1] * 10000] * 1000] * 10))
        repo.git_add(auto_lfs_track=True)
        repo.git_commit()
        _, result = repo.git_push(blocking=False)

        result._process.kill()

        while result._process.poll() is None:
            time.sleep(0.5)

        self.assertTrue(result.is_done)
        self.assertEqual(result.status, -9)

    def test_commit_context_manager(self):
        # Clone and commit from a first folder
        folder_1 = self.repo_path / "folder_1"
        clone = self.clone_repo(local_dir=folder_1)
        with clone.commit("Commit"):
            with open("dummy.txt", "w") as f:
                f.write("hello")
            with open("model.bin", "w") as f:
                f.write("hello")

        # Clone in second folder. Check existence of committed files
        folder_2 = self.repo_path / "folder_2"
        self.clone_repo(local_dir=folder_2)
        files = os.listdir(folder_2)
        self.assertTrue("dummy.txt" in files)
        self.assertTrue("model.bin" in files)

    def test_clone_skip_lfs_files(self):
        # Upload LFS file
        self._api.upload_file(path_or_fileobj=b"Bin file", path_in_repo="file.bin", repo_id=self.repo_id)

        repo = self.clone_repo(skip_lfs_files=True)
        file_bin = self.repo_path / "file.bin"

        self.assertTrue(file_bin.read_text().startswith("version"))

        repo.git_pull(lfs=True)

        self.assertEqual(file_bin.read_text(), "Bin file")

    def test_commits_on_correct_branch(self):
        repo = self.clone_repo()
        branch = repo.current_branch
        repo.git_checkout("new-branch", create_branch_ok=True)
        repo.git_checkout(branch)

        with repo.commit("New commit"):
            with open("file.txt", "w+") as f:
                f.write("Ok")

        repo.git_checkout("new-branch")

        with repo.commit("New commit"):
            with open("new_file.txt", "w+") as f:
                f.write("Ok")

        with SoftTemporaryDirectory() as tmp:
            clone = self.clone_repo(local_dir=tmp)
            files = os.listdir(clone.local_dir)
            self.assertTrue("file.txt" in files)
            self.assertFalse("new_file.txt" in files)

            clone.git_checkout("new-branch")
            files = os.listdir(clone.local_dir)
            self.assertFalse("file.txt" in files)
            self.assertTrue("new_file.txt" in files)

    def test_repo_checkout_push(self):
        repo = self.clone_repo()

        repo.git_checkout("new-branch", create_branch_ok=True)
        repo.git_checkout("main")

        (self.repo_path / "file.txt").write_text("OK")

        repo.push_to_hub("Commit #1")
        repo.git_checkout("new-branch", create_branch_ok=True)

        (self.repo_path / "new_file.txt").write_text("OK")

        repo.push_to_hub("Commit #2")

        with SoftTemporaryDirectory() as tmp:
            clone = self.clone_repo(local_dir=tmp)
            files = os.listdir(clone.local_dir)
            self.assertTrue("file.txt" in files)
            self.assertFalse("new_file.txt" in files)

            clone.git_checkout("new-branch")
            files = os.listdir(clone.local_dir)
            self.assertFalse("file.txt" in files)
            self.assertTrue("new_file.txt" in files)

    def test_repo_checkout_commit_context_manager(self):
        repo = self.clone_repo()

        with repo.commit("Commit #1", branch="new-branch"):
            with open(os.path.join(repo.local_dir, "file.txt"), "w+") as f:
                f.write("Ok")

        with repo.commit("Commit #2", branch="main"):
            with open(os.path.join(repo.local_dir, "new_file.txt"), "w+") as f:
                f.write("Ok")

        # Maintains lastly used branch
        with repo.commit("Commit #3"):
            with open(os.path.join(repo.local_dir, "new_file-2.txt"), "w+") as f:
                f.write("Ok")

        with SoftTemporaryDirectory() as tmp:
            clone = self.clone_repo(local_dir=tmp)
            files = os.listdir(clone.local_dir)
            self.assertFalse("file.txt" in files)
            self.assertTrue("new_file-2.txt" in files)
            self.assertTrue("new_file.txt" in files)

            clone.git_checkout("new-branch")
            files = os.listdir(clone.local_dir)
            self.assertTrue("file.txt" in files)
            self.assertFalse("new_file.txt" in files)
            self.assertFalse("new_file-2.txt" in files)

    def test_add_tag(self):
        repo = self.clone_repo()
        repo.add_tag("v4.6.0", remote="origin")
        self.assertTrue(repo.tag_exists("v4.6.0", remote="origin"))

    def test_add_annotated_tag(self):
        repo = self.clone_repo()
        repo.add_tag("v4.5.0", message="This is an annotated tag", remote="origin")

        # Unfortunately git offers no built-in way to check the annotated
        # message of a remote tag.
        # In order to check that the remote tag was correctly annotated,
        # we delete the local tag before pulling the remote tag (which
        # should be the same). We then check that this tag is correctly
        # annotated.
        repo.delete_tag("v4.5.0")

        self.assertTrue(repo.tag_exists("v4.5.0", remote="origin"))
        self.assertFalse(repo.tag_exists("v4.5.0"))

        # Tag still exists on remote
        run_subprocess("git pull --tags", folder=self.repo_path)
        self.assertTrue(repo.tag_exists("v4.5.0"))

        # Tag is annotated
        result = run_subprocess("git tag -n9", folder=self.repo_path).stdout.strip()
        self.assertIn("This is an annotated tag", result)

    def test_delete_tag(self):
        repo = self.clone_repo()

        repo.add_tag("v4.6.0", message="This is an annotated tag", remote="origin")
        self.assertTrue(repo.tag_exists("v4.6.0", remote="origin"))

        repo.delete_tag("v4.6.0")
        self.assertFalse(repo.tag_exists("v4.6.0"))
        self.assertTrue(repo.tag_exists("v4.6.0", remote="origin"))

        repo.delete_tag("v4.6.0", remote="origin")
        self.assertFalse(repo.tag_exists("v4.6.0", remote="origin"))

    def test_lfs_prune(self):
        repo = self.clone_repo()

        with repo.commit("Committing LFS file"):
            with open("file.bin", "w+") as f:
                f.write("Random string 1")

        with repo.commit("Committing LFS file"):
            with open("file.bin", "w+") as f:
                f.write("Random string 2")

        root_directory = self.repo_path / ".git" / "lfs"
        git_lfs_files_size = sum(f.stat().st_size for f in root_directory.glob("**/*") if f.is_file())
        repo.lfs_prune()
        post_prune_git_lfs_files_size = sum(f.stat().st_size for f in root_directory.glob("**/*") if f.is_file())

        # Size of the directory holding LFS files was reduced
        self.assertLess(post_prune_git_lfs_files_size, git_lfs_files_size)

    def test_lfs_prune_git_push(self):
        repo = self.clone_repo()
        with repo.commit("Committing LFS file"):
            with open("file.bin", "w+") as f:
                f.write("Random string 1")

        root_directory = self.repo_path / ".git" / "lfs"
        git_lfs_files_size = sum(f.stat().st_size for f in root_directory.glob("**/*") if f.is_file())

        with open(os.path.join(repo.local_dir, "file.bin"), "w+") as f:
            f.write("Random string 2")

        repo.git_add()
        repo.git_commit("New commit")
        repo.git_push(auto_lfs_prune=True)

        post_prune_git_lfs_files_size = sum(f.stat().st_size for f in root_directory.glob("**/*") if f.is_file())

        # Size of the directory holding LFS files is the exact same
        self.assertEqual(post_prune_git_lfs_files_size, git_lfs_files_size)


class TestRepositoryOffline(RepositoryTestAbstract):
    """Class to test `Repository` object on local folders only (no cloning from Hub)."""

    repo: Repository

    @classmethod
    @expect_deprecation("Repository")
    def setUp(self) -> None:
        super().setUp()

        run_subprocess("git init", folder=self.repo_path)

        self.repo = Repository(self.repo_path, git_user="ci", git_email="ci@dummy.ci")

        git_attributes_path = self.repo_path / ".gitattributes"
        git_attributes_path.write_text("*.pt filter=lfs diff=lfs merge=lfs -text")

        self.repo.git_add(".gitattributes")
        self.repo.git_commit("Add .gitattributes")

    def test_is_tracked_with_lfs(self):
        txt_1 = self.repo_path / "small_file_1.txt"
        txt_2 = self.repo_path / "small_file_2.txt"
        pt_1 = self.repo_path / "model.pt"

        txt_1.write_text(self.small_content)
        txt_2.write_text(self.small_content)
        pt_1.write_text(self.small_content)

        self.repo.lfs_track("small_file_1.txt")

        self.assertTrue(is_tracked_with_lfs(txt_1))
        self.assertFalse(is_tracked_with_lfs(txt_2))
        self.assertTrue(pt_1)

    def test_is_tracked_with_lfs_with_pattern(self):
        txt_small_file = self.repo_path / "small_file.txt"
        txt_small_file.write_text(self.small_content)

        txt_large_file = self.repo_path / "large_file.txt"
        txt_large_file.write_text(self.large_content)

        (self.repo_path / "dir").mkdir()
        txt_small_file_in_dir = self.repo_path / "dir" / "small_file.txt"
        txt_small_file_in_dir.write_text(self.small_content)

        txt_large_file_in_dir = self.repo_path / "dir" / "large_file.txt"
        txt_large_file_in_dir.write_text(self.large_content)

        self.repo.auto_track_large_files("dir")

        self.assertFalse(is_tracked_with_lfs(txt_large_file))
        self.assertFalse(is_tracked_with_lfs(txt_small_file))
        self.assertTrue(is_tracked_with_lfs(txt_large_file_in_dir))
        self.assertFalse(is_tracked_with_lfs(txt_small_file_in_dir))

    def test_auto_track_large_files(self):
        txt_small_file = self.repo_path / "small_file.txt"
        txt_small_file.write_text(self.small_content)

        txt_large_file = self.repo_path / "large_file.txt"
        txt_large_file.write_text(self.large_content)

        self.repo.auto_track_large_files()

        self.assertTrue(is_tracked_with_lfs(txt_large_file))
        self.assertFalse(is_tracked_with_lfs(txt_small_file))

    def test_auto_track_binary_files(self):
        non_binary_file = self.repo_path / "non_binary_file.txt"
        non_binary_file.write_text(self.small_content)

        binary_file = self.repo_path / "binary_file.txt"
        binary_file.write_text(self.binary_content)

        self.repo.auto_track_binary_files()

        self.assertFalse(is_tracked_with_lfs(non_binary_file))
        self.assertTrue(is_tracked_with_lfs(binary_file))

    def test_auto_track_large_files_ignored_with_gitignore(self):
        (self.repo_path / "dir").mkdir()

        # Test nested gitignores
        gitignore_file = self.repo_path / ".gitignore"
        gitignore_file.write_text("large_file.txt")

        gitignore_file_in_dir = self.repo_path / "dir" / ".gitignore"
        gitignore_file_in_dir.write_text("large_file_3.txt")

        large_file = self.repo_path / "large_file.txt"
        large_file.write_text(self.large_content)

        large_file_2 = self.repo_path / "large_file_2.txt"
        large_file_2.write_text(self.large_content)

        large_file_3 = self.repo_path / "dir" / "large_file_3.txt"
        large_file_3.write_text(self.large_content)

        large_file_4 = self.repo_path / "dir" / "large_file_4.txt"
        large_file_4.write_text(self.large_content)

        self.repo.auto_track_large_files()

        # Large files
        self.assertFalse(is_tracked_with_lfs(large_file))
        self.assertTrue(is_tracked_with_lfs(large_file_2))

        self.assertFalse(is_tracked_with_lfs(large_file_3))
        self.assertTrue(is_tracked_with_lfs(large_file_4))

    def test_auto_track_binary_files_ignored_with_gitignore(self):
        (self.repo_path / "dir").mkdir()

        # Test nested gitignores
        gitignore_file = self.repo_path / ".gitignore"
        gitignore_file.write_text("binary_file.txt")

        gitignore_file_in_dir = self.repo_path / "dir" / ".gitignore"
        gitignore_file_in_dir.write_text("binary_file_3.txt")

        binary_file = self.repo_path / "binary_file.txt"
        binary_file.write_text(self.binary_content)

        binary_file_2 = self.repo_path / "binary_file_2.txt"
        binary_file_2.write_text(self.binary_content)

        binary_file_3 = self.repo_path / "dir" / "binary_file_3.txt"
        binary_file_3.write_text(self.binary_content)

        binary_file_4 = self.repo_path / "dir" / "binary_file_4.txt"
        binary_file_4.write_text(self.binary_content)

        self.repo.auto_track_binary_files()

        # Binary files
        self.assertFalse(is_tracked_with_lfs(binary_file))
        self.assertTrue(is_tracked_with_lfs(binary_file_2))
        self.assertFalse(is_tracked_with_lfs(binary_file_3))
        self.assertTrue(is_tracked_with_lfs(binary_file_4))

    def test_auto_track_large_files_through_git_add(self):
        txt_small_file = self.repo_path / "small_file.txt"
        txt_small_file.write_text(self.small_content)

        txt_large_file = self.repo_path / "large_file.txt"
        txt_large_file.write_text(self.large_content)

        self.repo.git_add(auto_lfs_track=True)

        self.assertTrue(is_tracked_with_lfs(txt_large_file))
        self.assertFalse(is_tracked_with_lfs(txt_small_file))

    def test_auto_track_binary_files_through_git_add(self):
        non_binary_file = self.repo_path / "small_file.txt"
        non_binary_file.write_text(self.small_content)

        binary_file = self.repo_path / "binary.txt"
        binary_file.write_text(self.binary_content)

        self.repo.git_add(auto_lfs_track=True)

        self.assertTrue(is_tracked_with_lfs(binary_file))
        self.assertFalse(is_tracked_with_lfs(non_binary_file))

    def test_auto_no_track_large_files_through_git_add(self):
        txt_small_file = self.repo_path / "small_file.txt"
        txt_small_file.write_text(self.small_content)

        txt_large_file = self.repo_path / "large_file.txt"
        txt_large_file.write_text(self.large_content)

        self.repo.git_add(auto_lfs_track=False)

        self.assertFalse(is_tracked_with_lfs(txt_large_file))
        self.assertFalse(is_tracked_with_lfs(txt_small_file))

    def test_auto_no_track_binary_files_through_git_add(self):
        non_binary_file = self.repo_path / "small_file.txt"
        non_binary_file.write_text(self.small_content)

        binary_file = self.repo_path / "binary.txt"
        binary_file.write_text(self.binary_content)

        self.repo.git_add(auto_lfs_track=False)

        self.assertFalse(is_tracked_with_lfs(binary_file))
        self.assertFalse(is_tracked_with_lfs(non_binary_file))

    def test_auto_track_updates_removed_gitattributes(self):
        txt_small_file = self.repo_path / "small_file.txt"
        txt_small_file.write_text(self.small_content)

        txt_large_file = self.repo_path / "large_file.txt"
        txt_large_file.write_text(self.large_content)

        self.repo.git_add(auto_lfs_track=True)

        self.assertTrue(is_tracked_with_lfs(txt_large_file))
        self.assertFalse(is_tracked_with_lfs(txt_small_file))

        # Remove large file
        txt_large_file.unlink()

        # Auto track should remove the entry from .gitattributes
        self.repo.auto_track_large_files()

        # Recreate the large file with smaller contents
        txt_large_file.write_text(self.small_content)

        # Ensure the file is not LFS tracked anymore
        self.repo.auto_track_large_files()
        self.assertFalse(is_tracked_with_lfs(txt_large_file))

    def test_checkout_non_existing_branch(self):
        self.assertRaises(EnvironmentError, self.repo.git_checkout, "brand-new-branch")

    def test_checkout_new_branch(self):
        self.repo.git_checkout("new-branch", create_branch_ok=True)
        self.assertEqual(self.repo.current_branch, "new-branch")

    def test_is_not_tracked_upstream(self):
        self.repo.git_checkout("new-branch", create_branch_ok=True)
        self.assertFalse(is_tracked_upstream(self.repo.local_dir))

    def test_no_branch_checked_out_raises(self):
        head_commit_ref = run_subprocess("git show --oneline -s", folder=self.repo_path).stdout.split()[0]

        self.repo.git_checkout(head_commit_ref)
        self.assertRaises(OSError, is_tracked_upstream, self.repo.local_dir)

    @expect_deprecation("Repository")
    def test_repo_init_checkout_default_revision(self):
        # Instantiate repository on a given revision
        repo = Repository(self.repo_path, revision="new-branch")
        self.assertEqual(repo.current_branch, "new-branch")

        # The revision should be kept when re-initializing the repo
        repo_2 = Repository(self.repo_path)
        self.assertEqual(repo_2.current_branch, "new-branch")

    @expect_deprecation("Repository")
    def test_repo_init_checkout_revision(self):
        current_head_hash = self.repo.git_head_hash()

        (self.repo_path / "file.txt").write_text("hello world")

        self.repo.git_add()
        self.repo.git_commit("Add file.txt")

        new_head_hash = self.repo.git_head_hash()

        self.assertNotEqual(current_head_hash, new_head_hash)

        previous_head_repo = Repository(self.repo_path, revision=current_head_hash)
        files = os.listdir(previous_head_repo.local_dir)
        self.assertNotIn("file.txt", files)

        current_head_repo = Repository(self.repo_path, revision=new_head_hash)
        files = os.listdir(current_head_repo.local_dir)
        self.assertIn("file.txt", files)

    @expect_deprecation("Repository")
    def test_repo_user(self):
        _ = Repository(self.repo_path, token=TOKEN)
        username = run_subprocess("git config user.name", folder=self.repo_path).stdout
        email = run_subprocess("git config user.email", folder=self.repo_path).stdout

        # hardcode values to avoid another api call to whoami
        self.assertEqual(username.strip(), "Dummy User")
        self.assertEqual(email.strip(), "julien@huggingface.co")

    @expect_deprecation("Repository")
    def test_repo_passed_user(self):
        _ = Repository(self.repo_path, token=TOKEN, git_user="RANDOM_USER", git_email="EMAIL@EMAIL.EMAIL")
        username = run_subprocess("git config user.name", folder=self.repo_path).stdout
        email = run_subprocess("git config user.email", folder=self.repo_path).stdout

        self.assertEqual(username.strip(), "RANDOM_USER")
        self.assertEqual(email.strip(), "EMAIL@EMAIL.EMAIL")

    def test_add_tag(self):
        self.repo.add_tag("v4.6.0")
        self.assertTrue(self.repo.tag_exists("v4.6.0"))

    def test_add_annotated_tag(self):
        self.repo.add_tag("v4.6.0", message="This is an annotated tag")
        self.assertTrue(self.repo.tag_exists("v4.6.0"))

        result = run_subprocess("git tag -n9", folder=self.repo_path).stdout.strip()
        self.assertIn("This is an annotated tag", result)

    def test_delete_tag(self):
        self.repo.add_tag("v4.6.0", message="This is an annotated tag")
        self.assertTrue(self.repo.tag_exists("v4.6.0"))

        self.repo.delete_tag("v4.6.0")
        self.assertFalse(self.repo.tag_exists("v4.6.0"))

    def test_repo_clean(self):
        self.assertTrue(self.repo.is_repo_clean())
        (self.repo_path / "file.txt").write_text("hello world")
        self.assertFalse(self.repo.is_repo_clean())


class TestRepositoryDataset(RepositoryTestAbstract):
    """Class to test that cloning from a different repo_type works fine."""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.repo_url = cls._api.create_repo(repo_id=repo_name(), repo_type="dataset")
        cls.repo_id = cls.repo_url.repo_id
        cls._api.upload_file(
            path_or_fileobj=cls.binary_content.encode(),
            path_in_repo="file.txt",
            repo_id=cls.repo_id,
            repo_type="dataset",
        )

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls._api.delete_repo(repo_id=cls.repo_id, repo_type="dataset")

    @expect_deprecation("Repository")
    def test_clone_dataset_with_endpoint_explicit_repo_type(self):
        Repository(
            self.repo_path, clone_from=self.repo_url, repo_type="dataset", git_user="ci", git_email="ci@dummy.com"
        )
        self.assertTrue((self.repo_path / "file.txt").exists())

    @expect_deprecation("Repository")
    def test_clone_dataset_with_endpoint_implicit_repo_type(self):
        self.assertIn("dataset", self.repo_url)  # Implicit
        Repository(self.repo_path, clone_from=self.repo_url, git_user="ci", git_email="ci@dummy.com")
        self.assertTrue((self.repo_path / "file.txt").exists())

    @expect_deprecation("Repository")
    def test_clone_dataset_with_repo_id_and_repo_type(self):
        Repository(
            self.repo_path, clone_from=self.repo_id, repo_type="dataset", git_user="ci", git_email="ci@dummy.com"
        )
        self.assertTrue((self.repo_path / "file.txt").exists())

    @expect_deprecation("Repository")
    def test_clone_dataset_no_ci_user_and_email(self):
        Repository(self.repo_path, clone_from=self.repo_id, repo_type="dataset")
        self.assertTrue((self.repo_path / "file.txt").exists())

    @expect_deprecation("Repository")
    def test_clone_dataset_with_repo_name_and_repo_type_fails(self):
        with self.assertRaises(EnvironmentError):
            Repository(
                self.repo_path,
                clone_from=self.repo_id.split("/")[1],
                repo_type="dataset",
                token=TOKEN,
                git_user="ci",
                git_email="ci@dummy.com",
            )



# Copyright 2025 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains commands to interact with repositories on the Hugging Face Hub.

Usage:
    # create a new dataset repo on the Hub
    hf repo create my-cool-dataset --repo-type=dataset

    # create a private model repo on the Hub
    hf repo create my-cool-model --private
"""

import argparse
from argparse import _SubParsersAction
from typing import Optional

from requests.exceptions import HTTPError

from huggingface_hub.commands import BaseHuggingfaceCLICommand
from huggingface_hub.commands._cli_utils import ANSI
from huggingface_hub.constants import REPO_TYPES, SPACES_SDK_TYPES
from huggingface_hub.errors import HfHubHTTPError, RepositoryNotFoundError, RevisionNotFoundError
from huggingface_hub.hf_api import HfApi
from huggingface_hub.utils import logging


logger = logging.get_logger(__name__)


class RepoCommands(BaseHuggingfaceCLICommand):
    @staticmethod
    def register_subcommand(parser: _SubParsersAction):
        repo_parser = parser.add_parser("repo", help="Manage repos on the Hub.")
        repo_subparsers = repo_parser.add_subparsers(help="huggingface.co repos related commands")

        # Show help if no subcommand is provided
        repo_parser.set_defaults(func=lambda args: repo_parser.print_help())

        # CREATE
        repo_create_parser = repo_subparsers.add_parser("create", help="Create a new repo on huggingface.co")
        repo_create_parser.add_argument(
            "repo_id",
            type=str,
            help="The ID of the repo to create to (e.g. `username/repo-name`). The username is optional and will be set to your username if not provided.",
        )
        repo_create_parser.add_argument(
            "--repo-type",
            type=str,
            help='Optional: set to "dataset" or "space" if creating a dataset or space, default is model.',
        )
        repo_create_parser.add_argument(
            "--space_sdk",
            type=str,
            help='Optional: Hugging Face Spaces SDK type. Required when --type is set to "space".',
            choices=SPACES_SDK_TYPES,
        )
        repo_create_parser.add_argument(
            "--private",
            action="store_true",
            help="Whether to create a private repository. Defaults to public unless the organization's default is private.",
        )
        repo_create_parser.add_argument(
            "--token",
            type=str,
            help="Hugging Face token. Will default to the locally saved token if not provided.",
        )
        repo_create_parser.add_argument(
            "--exist-ok",
            action="store_true",
            help="Do not raise an error if repo already exists.",
        )
        repo_create_parser.add_argument(
            "--resource-group-id",
            type=str,
            help="Resource group in which to create the repo. Resource groups is only available for Enterprise Hub organizations.",
        )
        repo_create_parser.set_defaults(func=lambda args: RepoCreateCommand(args))

        # TAG SUBCOMMANDS
        repo_tag_parser = repo_subparsers.add_parser("tag", help="Manage tags for a repo on the Hub.")
        tag_subparsers = repo_tag_parser.add_subparsers(help="Tag actions", dest="tag_action", required=True)

        # tag create
        tag_create_parser = tag_subparsers.add_parser("create", help="Create a tag for a repo.")
        tag_create_parser.add_argument(
            "repo_id", type=str, help="The ID of the repo to tag (e.g. `username/repo-name`)."
        )
        tag_create_parser.add_argument("tag", type=str, help="The name of the tag to create.")
        tag_create_parser.add_argument("-m", "--message", type=str, help="The description of the tag to create.")
        tag_create_parser.add_argument("--revision", type=str, help="The git revision to tag.")
        tag_create_parser.add_argument(
            "--token", type=str, help="A User Access Token generated from https://huggingface.co/settings/tokens."
        )
        tag_create_parser.add_argument(
            "--repo-type",
            choices=["model", "dataset", "space"],
            default="model",
            help="Set the type of repository (model, dataset, or space).",
        )
        tag_create_parser.set_defaults(func=lambda args: RepoTagCreateCommand(args))

        # tag list
        tag_list_parser = tag_subparsers.add_parser("list", help="List tags for a repo.")
        tag_list_parser.add_argument(
            "repo_id", type=str, help="The ID of the repo to list tags for (e.g. `username/repo-name`)."
        )
        tag_list_parser.add_argument(
            "--token", type=str, help="A User Access Token generated from https://huggingface.co/settings/tokens."
        )
        tag_list_parser.add_argument(
            "--repo-type",
            choices=["model", "dataset", "space"],
            default="model",
            help="Set the type of repository (model, dataset, or space).",
        )
        tag_list_parser.set_defaults(func=lambda args: RepoTagListCommand(args))

        # tag delete
        tag_delete_parser = tag_subparsers.add_parser("delete", help="Delete a tag from a repo.")
        tag_delete_parser.add_argument(
            "repo_id", type=str, help="The ID of the repo to delete the tag from (e.g. `username/repo-name`)."
        )
        tag_delete_parser.add_argument("tag", type=str, help="The name of the tag to delete.")
        tag_delete_parser.add_argument("-y", "--yes", action="store_true", help="Answer Yes to prompts automatically.")
        tag_delete_parser.add_argument(
            "--token", type=str, help="A User Access Token generated from https://huggingface.co/settings/tokens."
        )
        tag_delete_parser.add_argument(
            "--repo-type",
            choices=["model", "dataset", "space"],
            default="model",
            help="Set the type of repository (model, dataset, or space).",
        )
        tag_delete_parser.set_defaults(func=lambda args: RepoTagDeleteCommand(args))


class RepoCreateCommand:
    def __init__(self, args: argparse.Namespace):
        self.repo_id: str = args.repo_id
        self.repo_type: Optional[str] = args.repo_type
        self.space_sdk: Optional[str] = args.space_sdk
        self.private: bool = args.private
        self.token: Optional[str] = args.token
        self.exist_ok: bool = args.exist_ok
        self.resource_group_id: Optional[str] = args.resource_group_id
        self._api = HfApi()

    def run(self):
        repo_url = self._api.create_repo(
            repo_id=self.repo_id,
            repo_type=self.repo_type,
            private=self.private,
            token=self.token,
            exist_ok=self.exist_ok,
            resource_group_id=self.resource_group_id,
            space_sdk=self.space_sdk,
        )
        print(f"Successfully created {ANSI.bold(repo_url.repo_id)} on the Hub.")
        print(f"Your repo is now available at {ANSI.bold(repo_url)}")


class RepoTagCommand:
    def __init__(self, args):
        self.args = args
        self.api = HfApi(token=getattr(args, "token", None))
        self.repo_id = args.repo_id
        self.repo_type = getattr(args, "repo_type", "model")
        if self.repo_type not in REPO_TYPES:
            print("Invalid repo --repo-type")
            exit(1)


class RepoTagCreateCommand(RepoTagCommand):
    def run(self):
        print(
            f"You are about to create tag {ANSI.bold(str(self.args.tag))} on {self.repo_type} {ANSI.bold(self.repo_id)}"
        )
        try:
            self.api.create_tag(
                repo_id=self.repo_id,
                tag=self.args.tag,
                tag_message=getattr(self.args, "message", None),
                revision=getattr(self.args, "revision", None),
                repo_type=self.repo_type,
            )
        except RepositoryNotFoundError:
            print(f"{self.repo_type.capitalize()} {ANSI.bold(self.repo_id)} not found.")
            exit(1)
        except RevisionNotFoundError:
            print(f"Revision {ANSI.bold(str(getattr(self.args, 'revision', None)))} not found.")
            exit(1)
        except HfHubHTTPError as e:
            if e.response.status_code == 409:
                print(f"Tag {ANSI.bold(str(self.args.tag))} already exists on {ANSI.bold(self.repo_id)}")
                exit(1)
            raise e
        print(f"Tag {ANSI.bold(str(self.args.tag))} created on {ANSI.bold(self.repo_id)}")


class RepoTagListCommand(RepoTagCommand):
    def run(self):
        try:
            refs = self.api.list_repo_refs(
                repo_id=self.repo_id,
                repo_type=self.repo_type,
            )
        except RepositoryNotFoundError:
            print(f"{self.repo_type.capitalize()} {ANSI.bold(self.repo_id)} not found.")
            exit(1)
        except HTTPError as e:
            print(e)
            print(ANSI.red(e.response.text))
            exit(1)
        if len(refs.tags) == 0:
            print("No tags found")
            exit(0)
        print(f"Tags for {self.repo_type} {ANSI.bold(self.repo_id)}:")
        for tag in refs.tags:
            print(tag.name)


class RepoTagDeleteCommand(RepoTagCommand):
    def run(self):
        print(f"You are about to delete tag {ANSI.bold(self.args.tag)} on {self.repo_type} {ANSI.bold(self.repo_id)}")
        if not getattr(self.args, "yes", False):
            choice = input("Proceed? [Y/n] ").lower()
            if choice not in ("", "y", "yes"):
                print("Abort")
                exit()
        try:
            self.api.delete_tag(repo_id=self.repo_id, tag=self.args.tag, repo_type=self.repo_type)
        except RepositoryNotFoundError:
            print(f"{self.repo_type.capitalize()} {ANSI.bold(self.repo_id)} not found.")
            exit(1)
        except RevisionNotFoundError:
            print(f"Tag {ANSI.bold(self.args.tag)} not found on {ANSI.bold(self.repo_id)}")
            exit(1)
        print(f"Tag {ANSI.bold(self.args.tag)} deleted on {ANSI.bold(self.repo_id)}")



# Inference code generated from the JSON schema spec in @huggingface/tasks.
#
# See:
#   - script: https://github.com/huggingface/huggingface.js/blob/main/packages/tasks/scripts/inference-codegen.ts
#   - specs:  https://github.com/huggingface/huggingface.js/tree/main/packages/tasks/src/tasks.
from typing import Any, Dict, Literal, Optional

from .base import BaseInferenceType, dataclass_with_extra


TranslationTruncationStrategy = Literal["do_not_truncate", "longest_first", "only_first", "only_second"]


@dataclass_with_extra
class TranslationParameters(BaseInferenceType):
    """Additional inference parameters for Translation"""

    clean_up_tokenization_spaces: Optional[bool] = None
    """Whether to clean up the potential extra spaces in the text output."""
    generate_parameters: Optional[Dict[str, Any]] = None
    """Additional parametrization of the text generation algorithm."""
    src_lang: Optional[str] = None
    """The source language of the text. Required for models that can translate from multiple
    languages.
    """
    tgt_lang: Optional[str] = None
    """Target language to translate to. Required for models that can translate to multiple
    languages.
    """
    truncation: Optional["TranslationTruncationStrategy"] = None
    """The truncation strategy to use."""


@dataclass_with_extra
class TranslationInput(BaseInferenceType):
    """Inputs for Translation inference"""

    inputs: str
    """The text to translate."""
    parameters: Optional[TranslationParameters] = None
    """Additional inference parameters for Translation"""


@dataclass_with_extra
class TranslationOutput(BaseInferenceType):
    """Outputs of inference for the Translation task"""

    translation_text: str
    """The translated text."""



# Copyright 2020 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contain tests for `HfFolder` utility."""

import os
import unittest
from uuid import uuid4

from huggingface_hub.utils import HfFolder


def _generate_token() -> str:
    return f"token-{uuid4()}"


class HfFolderTest(unittest.TestCase):
    def test_token_workflow(self):
        """
        Test the whole token save/get/delete workflow,
        with the desired behavior with respect to non-existent tokens.
        """
        token = _generate_token()
        HfFolder.save_token(token)
        self.assertEqual(HfFolder.get_token(), token)
        HfFolder.delete_token()
        HfFolder.delete_token()
        # ^^ not an error, we test that the
        # second call does not fail.
        self.assertEqual(HfFolder.get_token(), None)
        # test TOKEN in env
        self.assertEqual(HfFolder.get_token(), None)
        with unittest.mock.patch.dict(os.environ, {"HF_TOKEN": token}):
            self.assertEqual(HfFolder.get_token(), token)

    def test_token_strip(self):
        """
        Test the workflow when the token is mistakenly finishing with new-line or space character.
        """
        token = _generate_token()
        HfFolder.save_token(" " + token + "\n")
        self.assertEqual(HfFolder.get_token(), token)
        HfFolder.delete_token()
